//	Altirra - Atari 800/800XL/5200 emulator
//	Copyright (C) 2009-2020 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License along
//	with this program. If not, see <http://www.gnu.org/licenses/>.

#ifndef f_AT_CUSTOMDEVICECOMPILER_H
#define f_AT_CUSTOMDEVICECOMPILER_H

#include <vd2/system/vdtypes.h>
#include "customdevicevm.h"

class ATCDVMDomain;
struct ATCDVMFunction;

class ATCustomDeviceCompiler {
public:
	ATCustomDeviceCompiler(ATCDVMDomain& domain);

	bool IsValidVariableName(const char *name);
	bool DefineObjectVariable(const char *name, ATCDVMObject *obj, const ATCDVMObjectClass& objClass);
	bool DefineSpecialObjectVariable(const char *name, ATCDVMObject *obj, const ATCDVMObjectClass& objClass);

	template<typename T>
	bool DefineObjectVariable(const char *name, T *obj) {
		static_assert(std::is_convertible_v<T, ATCDVMObject>);
		return DefineObjectVariable(name, obj, T::kVMObjectClass);
	}

	template<typename T>
	bool DefineSpecialObjectVariable(const char *name, T *obj) {
		static_assert(std::is_convertible_v<T, ATCDVMObject>);
		return DefineSpecialObjectVariable(name, obj, T::kVMObjectClass);
	}

	bool DefineIntegerVariable(const char *name);
	void DefineSpecialVariable(const char *name);
	void DefineThreadVariable(const char *name);

	const ATCDVMTypeInfo *GetVariable(const char *name) const;

	const ATCDVMFunction *CompileProcedure(const ATCDVMTypeInfo& returnType, const char *src, size_t len);
	bool CompileFile(const char *src, size_t len);

	const char *GetError() const;
	size_t GetErrorOffset() const;
	std::pair<uint32, uint32> GetErrorLinePos() const;

private:
	enum Token : uint32 {
		kTokEnd = 0,
		kTokError = 256,
		kTokInteger,
		kTokIdentifier,
		kTokSpecialIdent,
		kTokLe,
		kTokGe,
		kTokNe,
		kTokEq,
		kTokLeftShift,
		kTokRightShift,
		kTokLogicalAnd,
		kTokLogicalOr,
		kTokIf,
		kTokReturn,
		kTokVoid,
		kTokInt,
		kTokFunction,
		kTokElse,
		kTokTrue,
		kTokFalse,
	};

	typedef ATCDVMTypeClass TypeClass;
	typedef ATCDVMTypeInfo TypeInfo;

	void InitSource(const void *src, size_t len);

	bool ParseFile();
	bool ParseVariableDefinition(uint32 tok);
	bool ParseFunction();
	ATCDVMFunction *ParseFunction(const ATCDVMTypeInfo& returnType);
	bool ParseBlock(bool& hasReturn);
	bool ParseStatement(uint32 tok, bool& hasReturn);
	bool ParseIfStatement(bool& hasReturn);
	bool ParseReturnStatement();
	bool ParseExpressionStatement(uint32 tok);
	bool ParseIntExpression();
	bool ParseExpression(TypeInfo& returnType);
	bool ParseLogicalOrExpression(TypeInfo& returnType);
	bool ParseLogicalAndExpression(TypeInfo& returnType);
	bool ParseEqualityExpression(TypeInfo& returnType);
	bool ParseRelationalExpression(TypeInfo& returnType);
	bool ParseBitwiseOrExpression(TypeInfo& returnType);
	bool ParseBitwiseXorExpression(TypeInfo& returnType);
	bool ParseBitwiseAndExpression(TypeInfo& returnType);
	bool ParseShiftExpression(TypeInfo& returnType);
	bool ParseAdditiveExpression(TypeInfo& returnType);
	bool ParseMultiplicativeExpression(TypeInfo& returnType);
	bool ParseUnaryExpression(TypeInfo& returnType);
	bool ParsePostfixExpression(TypeInfo& returnType);
	bool ParseAssignmentExpression(TypeInfo& returnType);
	bool ParseValue(TypeInfo& returnType);
	void ConvertToRvalue(TypeInfo& returnType);

	void Push(uint32 tok) { mPushedToken = tok; }
	uint32 Token();
	bool ReportError(const char *msg);
	bool ReportErrorF(const char *format, ...);

	const char *mpSrc = nullptr;
	const char *mpSrcStart = nullptr;
	const char *mpSrcEnd = nullptr;

	uint32 mPushedToken = 0;
	sint32 mTokValue = 0;
	VDStringSpanA mTokIdent;
	VDStringA mError;

	ATCDVMDomain *mpDomain = nullptr;
	vdfastvector<uint8> mByteCodeBuffer;
	vdfastvector<void(*)()> mMethodPtrs;
	const ATCDVMTypeInfo *mpReturnType;

	uint32 mVariableCount = 0;
	vdhashmap<VDStringA, TypeInfo, vdhash<VDStringA>, vdstringpred> mVariableLookup;
	vdhashmap<VDStringA, TypeInfo, vdhash<VDStringA>, vdstringpred> mSpecialVariableLookup;
	vdhashmap<VDStringA, TypeInfo, vdhash<VDStringA>, vdstringpred> mThreadVariableLookup;
	vdhashmap<VDStringA, uint32, vdhash<VDStringA>, vdstringpred> mFunctionLookup;
};

#endif
