//	Altirra - Atari 800/800XL/5200 emulator
//	Copyright (C) 2009-2020 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License along
//	with this program. If not, see <http://www.gnu.org/licenses/>.

#ifndef f_AT_CUSTOMDEVICEVM_H
#define f_AT_CUSTOMDEVICEVM_H

#include <type_traits>
#include <vd2/system/vdtypes.h>
#include <vd2/system/linearalloc.h>
#include <vd2/system/vdstl.h>

struct ATCDVMObject;

enum class ATCDVMOpcode : uint8 {
	Nop,
	Pop,
	Dup,
	IVLoad,			// varIdx.b
	IVStore,		// varIdx.b
	IALoad,
	IAStore,
	ISLoad,			// specialVarIdx.b
	ITLoad,			// threadVarIdx.b
	IntConst,		// const.l
	IntConst8,		// const.b
	IntAdd,
	IntSub,
	IntMul,
	IntDiv,
	IntMod,
	IntAnd,
	IntOr,
	IntXor,
	IntAsr,
	IntAsl,
	Not,
	And,
	Or,
	IntLt,
	IntLe,
	IntGt,
	IntGe,
	IntEq,
	IntNe,
	IntNeg,
	IntNot,
	Jz,
	Jnz,
	Jmp,
	Ljz,
	Ljnz,
	Ljmp,
	MethodCallVoid,		// argcount.b, methodindex.b
	MethodCallInt,		// argcount.b, methodindex.b
	FunctionCallVoid,	// argcount.b, methodindex.b
	FunctionCallInt,	// argcount.b, methodindex.b
	ReturnVoid,
	ReturnInt
};

uint32 ATCDVMGetOpcodeLength(ATCDVMOpcode opcode);

struct ATCDVMObjectClass;

enum class ATCDVMTypeClass : uint8 {
	Void,
	Int,
	IntLValue,
	Object,
	ObjectLValue
};

struct ATCDVMTypeInfo {
	ATCDVMTypeClass mClass;
	uint32 mIndex;
	const ATCDVMObjectClass *mpObjectClass;

	bool operator==(const ATCDVMTypeInfo& other) const {
		return mClass == other.mClass
			&& mIndex == other.mIndex
			&& mpObjectClass == other.mpObjectClass;
	}

	bool operator!=(const ATCDVMTypeInfo& other) const {
		return mClass != other.mClass
			|| mIndex != other.mIndex
			|| mpObjectClass != other.mpObjectClass;
	}
};

extern const ATCDVMTypeInfo kATCDVMTypeVoid;
extern const ATCDVMTypeInfo kATCDVMTypeInt;

template<typename T>
inline constexpr ATCDVMTypeInfo kATCDVMTypeClass {
	ATCDVMTypeClass::Object,
	0,
	&T::kVMObjectClass
};

template<typename T>
inline constexpr const ATCDVMTypeInfo& ATCDVMGetType() {
	return kATCDVMTypeClass<std::remove_pointer_t<T>>;
}

template<> inline constexpr const ATCDVMTypeInfo& ATCDVMGetType<void>() { return kATCDVMTypeVoid; }
template<> inline constexpr const ATCDVMTypeInfo& ATCDVMGetType<sint32>() { return kATCDVMTypeInt; }

template<typename T_Return, typename T_Object, typename... T_Args>
inline constexpr const ATCDVMTypeInfo *g_ATCDVMExternalMethodArgs[] = {
	&ATCDVMGetType<std::remove_cv_t<std::remove_reference_t<T_Return>>>(),
	&ATCDVMGetType<std::remove_cv_t<std::remove_reference_t<T_Args>>>()...
};

template<typename T> struct ATCDVMThunk;

template<typename T>
T ATCDVMDecodeThunkArgument(ATCDVMObject *const *objectTable, const sint32 *argPtr) {
	return static_cast<T>(objectTable[*argPtr]);
}

template<>
inline sint32 ATCDVMDecodeThunkArgument<sint32>(ATCDVMObject *const *objectTable, const sint32 *argPtr) {
	return *argPtr;
}

typedef void (*ATCDVMExternalVoidMethod)(ATCDVMObject *const *objectTable, const sint32 *args);
typedef sint32 (*ATCDVMExternalIntMethod)(ATCDVMObject *const *objectTable, const sint32 *args);

struct ATCDVMExternalMethod {
	template<auto T>
	static constexpr ATCDVMExternalMethod Bind(const char *name);

	const char *mpName;

	union {
		ATCDVMExternalVoidMethod mpVoidMethod;
		ATCDVMExternalIntMethod mpIntMethod;
	};

	const ATCDVMTypeInfo *const *mpTypes;
	uint32 mNumArgs;
};

template<typename T_Return, typename T_Object, typename... T_Args>
struct ATCDVMThunk<T_Return (T_Object::*)(T_Args...)> {
	typedef T_Return ReturnType;

	template<T_Return (T_Object::*T_Fn)(T_Args...)>
	static constexpr void GetPrototype(ATCDVMExternalMethod& method) {
		method.mpTypes = g_ATCDVMExternalMethodArgs<T_Return, T_Object, T_Args...>;
		method.mNumArgs = sizeof...(T_Args);
	}

	static constexpr uint32 kNumArgs = sizeof...(T_Args);

	template<T_Return (T_Object::*T_Fn)(T_Args...), uint32... T_ArgIndices>
	static constexpr auto GetThunkFunction(std::integer_sequence<uint32, T_ArgIndices...>) {
		return [](ATCDVMObject *const *objectTable, const sint32 *args) {
			return (static_cast<T_Object *>(objectTable[args[0]])->*T_Fn)(
				ATCDVMDecodeThunkArgument<T_Args>(objectTable, &args[T_ArgIndices + 1])...
			);
		};
	}

	template<T_Return (T_Object::*T_Fn)(T_Args...)>
	static constexpr auto GetThunkFunction() {
		return GetThunkFunction<T_Fn>(std::make_integer_sequence<uint32, kNumArgs>());
	}
};

struct ATCDVMObject {};

template<auto T>
constexpr ATCDVMExternalMethod ATCDVMExternalMethod::Bind(const char *name) {
	using ThunkType = ATCDVMThunk<decltype(T)>;

	ATCDVMExternalMethod method {};

	if constexpr (std::is_same_v<typename ThunkType::ReturnType, sint32>)
		method.mpIntMethod = ThunkType::template GetThunkFunction<T>();
	else
		method.mpVoidMethod = ThunkType::template GetThunkFunction<T>();

	ThunkType::GetPrototype<T>(method);

	method.mpName = name;

	return method;
}

struct ATCDVMObjectClass {
	const char *mpClassName;
	std::initializer_list<ATCDVMExternalMethod> mMethods;
};

struct ATCDVMFunction {
	const ATCDVMFunction *mpNext;
	const char *mpName;

	ATCDVMTypeInfo mReturnType;
	uint8 *mpByteCode;
	void (*const *mpMethodTable)();
	uint32 mByteCodeLen;
	uint32 mStackSlotsRequired;
};

class ATCDVMDomain {
public:
	VDLinearAllocator mAllocator;
	vdfastvector<sint32> mGlobalVariables;
	vdfastvector<sint32> mSpecialVariables;
	vdfastvector<ATCDVMObject *> mGlobalObjects;
	vdfastvector<const ATCDVMFunction *> mFunctions;
	uint32 mNumThreadVariables = 0;

	void Clear();
};

class ATCDVMThread {
public:
	struct StackFrame {
		const ATCDVMFunction *mpFunction;
		const uint8 *mpPC;
		uint32 mBP;
		uint32 mSP;
	};

	ATCDVMDomain *mpDomain;
	vdfastvector<StackFrame> mStackFrames;
	vdfastvector<sint32> mArgStack;
	vdfastvector<sint32> mThreadVariables;
	bool mbSuspend = false;

	void Init(ATCDVMDomain& domain);

	bool RunVoid(const ATCDVMFunction& function);
	sint32 RunInt(const ATCDVMFunction& function);

	void Suspend() { mbSuspend = true; }
	bool Resume();
	void Abort();

private:
	bool Run();
};

#endif
