//	Altirra - Atari 800/800XL/5200 emulator
//	Copyright (C) 2009-2020 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License along
//	with this program. If not, see <http://www.gnu.org/licenses/>.

#include <stdafx.h>
#include <vd2/system/binary.h>
#include <vd2/system/date.h>
#include <vd2/system/file.h>
#include <vd2/system/filesys.h>
#include <vd2/vdjson/jsonreader.h>
#include <at/atcore/propertyset.h>
#include <at/atcore/scheduler.h>
#include <at/atcore/vfs.h>
#include "customdevice.h"
#include "customdevicecompiler.h"
#include "customdevice_win32.h"
#include "inputcontroller.h"
#include "memorymanager.h"
#include "simulator.h"

extern ATSimulator g_sim;

class ATCustomDeviceDescException : public MyError {
public:
	ATCustomDeviceDescException(const VDJSONValueRef& context, const char *msg)
		: MyError("Error on line %u of custom device description: %s", context.GetLineNumber(), msg)
	{
	}

protected:
	ATCustomDeviceDescException() = default;

	void Format(const VDJSONValueRef& context, const char *format, va_list args) {
		VDStringA s;
		s.append_vsprintf(format, args);

		setf("Error on line %u of custom device description: %s", context.GetLineNumber(), s.c_str());
	}
};

class ATInvalidCustomDeviceDescException : public ATCustomDeviceDescException {
public:
	using ATCustomDeviceDescException::ATCustomDeviceDescException;
};

class ATInvalidCustomDeviceDescExceptionF : public ATCustomDeviceDescException {
public:
	ATInvalidCustomDeviceDescExceptionF(const VDJSONValueRef& context, const char *format, ...) {
		va_list args;

		va_start(args, format);
		Format(context, format, args);
		va_end(args);
	}
};

class ATUnsupportedCustomDeviceDescException : public ATCustomDeviceDescException {
public:
	using ATCustomDeviceDescException::ATCustomDeviceDescException;
};

void ATCreateDeviceCustom(const ATPropertySet& pset, IATDevice **dev) {
	vdrefptr<ATDeviceCustom> p(new ATDeviceCustom);

	*dev = p.release();
}

extern const ATDeviceDefinition g_ATDeviceDefCustom = { "custom", "custom", L"Custom", ATCreateDeviceCustom };

/////////////////////////////////////////////////////////////////////////////

const ATCDVMObjectClass ATDeviceCustom::Segment::kVMObjectClass {
	"Segment",
	{
		ATCDVMExternalMethod::Bind<&ATDeviceCustom::Segment::VMCallClear>("clear"),
		ATCDVMExternalMethod::Bind<&ATDeviceCustom::Segment::VMCallFill>("fill"),
		ATCDVMExternalMethod::Bind<&ATDeviceCustom::Segment::VMCallCopy>("copy"),
		ATCDVMExternalMethod::Bind<&ATDeviceCustom::Segment::VMCallReadByte>("read_byte"),
		ATCDVMExternalMethod::Bind<&ATDeviceCustom::Segment::VMCallWriteByte>("write_byte")
	}
};

const ATCDVMObjectClass ATDeviceCustom::MemoryLayer::kVMObjectClass {
	"MemoryLayer",
	{
		ATCDVMExternalMethod::Bind<&ATDeviceCustom::MemoryLayer::VMCallSetOffset>("set_offset"),
		ATCDVMExternalMethod::Bind<&ATDeviceCustom::MemoryLayer::VMCallSetSegmentAndOffset>("set_segment_and_offset"),
		ATCDVMExternalMethod::Bind<&ATDeviceCustom::MemoryLayer::VMCallSetModes>("set_modes"),
		ATCDVMExternalMethod::Bind<&ATDeviceCustom::MemoryLayer::VMCallSetReadOnly>("set_readonly")
	}
};

const ATCDVMObjectClass ATDeviceCustom::Network::kVMObjectClass {
	"Network",
	{
		ATCDVMExternalMethod::Bind<&ATDeviceCustom::Network::VMCallSendMessage>("send_message"),
	}
};

const ATCDVMObjectClass ATDeviceCustom::SIO::kVMObjectClass {
	"SIO",
	{
		ATCDVMExternalMethod::Bind<&ATDeviceCustom::SIO::VMCallAck>("ack"),
		ATCDVMExternalMethod::Bind<&ATDeviceCustom::SIO::VMCallNak>("nak"),
		ATCDVMExternalMethod::Bind<&ATDeviceCustom::SIO::VMCallError>("error"),
		ATCDVMExternalMethod::Bind<&ATDeviceCustom::SIO::VMCallComplete>("complete"),
		ATCDVMExternalMethod::Bind<&ATDeviceCustom::SIO::VMCallSendFrame>("send_frame"),
		ATCDVMExternalMethod::Bind<&ATDeviceCustom::SIO::VMCallRecvFrame>("recv_frame"),
		ATCDVMExternalMethod::Bind<&ATDeviceCustom::SIO::VMCallDelay>("delay"),
	}
};

const ATCDVMObjectClass ATDeviceCustom::Clock::kVMObjectClass {
	"Clock",
	{
		ATCDVMExternalMethod::Bind<&ATDeviceCustom::Clock::VMCallCaptureLocalTime>("capture_local_time"),
		ATCDVMExternalMethod::Bind<&ATDeviceCustom::Clock::VMCallLocalYear>("local_year"),
		ATCDVMExternalMethod::Bind<&ATDeviceCustom::Clock::VMCallLocalMonth>("local_month"),
		ATCDVMExternalMethod::Bind<&ATDeviceCustom::Clock::VMCallLocalDay>("local_day"),
		ATCDVMExternalMethod::Bind<&ATDeviceCustom::Clock::VMCallLocalDayOfWeek>("local_day_of_week"),
		ATCDVMExternalMethod::Bind<&ATDeviceCustom::Clock::VMCallLocalHour>("local_hour"),
		ATCDVMExternalMethod::Bind<&ATDeviceCustom::Clock::VMCallLocalMinute>("local_minute"),
		ATCDVMExternalMethod::Bind<&ATDeviceCustom::Clock::VMCallLocalSecond>("local_second"),
	}
};

const ATCDVMObjectClass ATDeviceCustom::ControllerPort::kVMObjectClass {
	"ControllerPort",
	{
		ATCDVMExternalMethod::Bind<&ATDeviceCustom::ControllerPort::VMCallSetPaddleA>("set_paddle_a"),
		ATCDVMExternalMethod::Bind<&ATDeviceCustom::ControllerPort::VMCallSetPaddleB>("set_paddle_b"),
		ATCDVMExternalMethod::Bind<&ATDeviceCustom::ControllerPort::VMCallSetTrigger>("set_trigger"),
		ATCDVMExternalMethod::Bind<&ATDeviceCustom::ControllerPort::VMCallSetDirs>("set_dirs"),
	}
};

/////////////////////////////////////////////////////////////////////////////

ATDeviceCustom::ATDeviceCustom() {
	mNetwork.mpParent = this;
	mSIO.mpParent = this;
	mSIOFrameSegment.mbReadOnly = true;
	mClock.mpParent = this;
}

void *ATDeviceCustom::AsInterface(uint32 iid) {
	switch(iid) {
		case IATDeviceMemMap::kTypeID: return static_cast<IATDeviceMemMap *>(this);
		case IATDeviceCartridge::kTypeID: return static_cast<IATDeviceCartridge *>(this);
		case IATDeviceIndicators::kTypeID: return static_cast<IATDeviceIndicators *>(this);
		case IATDeviceScheduling::kTypeID: return static_cast<IATDeviceScheduling *>(this);
		case IATDeviceSIO::kTypeID: return static_cast<IATDeviceSIO *>(this);
		default:
			return ATDevice::AsInterface(iid);
	}
}

void ATDeviceCustom::GetDeviceInfo(ATDeviceInfo& info) {
	info.mpDef = &g_ATDeviceDefCustom;
}

void ATDeviceCustom::GetSettingsBlurb(VDStringW& buf) {
	if (!mDeviceName.empty())
		buf += mDeviceName;
	else
		buf += mConfigPath;
}

void ATDeviceCustom::GetSettings(ATPropertySet& settings) {
	settings.SetString("path", mConfigPath.c_str());
	settings.SetBool("hotreload", mbHotReload);
}

bool ATDeviceCustom::SetSettings(const ATPropertySet& settings) {
	VDStringW newPath(settings.GetString("path", L""));
	const bool hotReload = settings.GetBool("hotreload");

	if (mConfigPath != newPath || mbHotReload != hotReload) {
		mConfigPath = newPath;
		mbHotReload = hotReload;

		if (mbInited)
			ReloadConfig();
	}

	return true;
}

void ATDeviceCustom::Init() {
	ATPortController *pca = g_sim.GetPortControllerA();
	ATPortController *pcb = g_sim.GetPortControllerB();

	const auto hardwareMode = g_sim.GetHardwareMode();
	const bool hasPort34 = (hardwareMode != kATHardwareMode_800XL && hardwareMode != kATHardwareMode_1200XL && hardwareMode != kATHardwareMode_XEGS && hardwareMode != kATHardwareMode_130XE);

	for(int i=0; i<4; ++i) {
		auto& port = mControllerPorts[i];
		port.mpPortController = i & 2 ? pcb : pca;
		port.mbPort2 = (i & 1) != 0;

		if (i < 2 || hasPort34)
			port.Enable();
		else
			port.Disable();
	}

	mbInited = true;

	ReloadConfig();

	if (mpNetworkEngine)
		mpNetworkEngine->WaitForFirstConnectionAttempt();
}

void ATDeviceCustom::Shutdown() {
	mbInited = false;

	ShutdownCustomDevice();

	mpIndicators = nullptr;
	mpCartPort = nullptr;
	mpMemMan = nullptr;
	mpScheduler = nullptr;
	mpSIOMgr = nullptr;
}

void ATDeviceCustom::ColdReset() {
	if (!mbInitedCustomDevice) {
		uint64 t = mpScheduler->GetTick64();

		if (mLastReloadAttempt != t) {
			mLastReloadAttempt = t;

			ReloadConfig();
		}
	}

	ReinitSegmentData(false);

	mClock.Reset();

	if (mpScriptEventColdReset)
		mVMThread.RunVoid(*mpScriptEventColdReset);

	if (mpNetworkEngine) {
		if (mpNetworkEngine->IsConnected())
			SendNetCommand(0, 0, NetworkCommand::ColdReset);
		else {
			uint32 t = ATSCHEDULER_GETTIME(mpScheduler);

			if (mLastConnectionErrorCycle != t) {
				mLastConnectionErrorCycle = t;

				mpIndicators->ReportError(L"No connection to device server. Custom device may not function properly.");
			}
		}
	}

	// reapply controller port state in case hardware mode has changed and we need to
	// set ports that were previously disabled (3/4)
	for(auto& port : mControllerPorts)
		port.Reapply();
}

void ATDeviceCustom::WarmReset() {	
	if (mpScriptEventWarmReset)
		mVMThread.RunVoid(*mpScriptEventWarmReset);

	if (mpNetworkEngine)
		SendNetCommand(0, 0, NetworkCommand::WarmReset);
}

bool ATDeviceCustom::GetErrorStatus(uint32 idx, VDStringW& error) {
	if (!mLastError.empty()) {
		if (!idx--) {
			error = mLastError;
			return true;
		}
	}

	if (mpNetworkEngine && !mpNetworkEngine->IsConnected()) {
		if (!idx--) {
			error = L"No connection to device server. Custom device may not function properly.";
			return true;
		}
	}

	return false;
}

void ATDeviceCustom::InitMemMap(ATMemoryManager *memmap) {
	mpMemMan = memmap;
}

bool ATDeviceCustom::GetMappedRange(uint32 index, uint32& lo, uint32& hi) const {
	return false;
}

void ATDeviceCustom::InitCartridge(IATDeviceCartridgePort *cartPort) {
	mpCartPort = cartPort;
}

bool ATDeviceCustom::IsLeftCartActive() const {
	for(const MemoryLayer& ml : mMemoryLayers) {
		if (ml.mbRD5Active)
			return true;
	}

	return false;
}

void ATDeviceCustom::SetCartEnables(bool leftEnable, bool rightEnable, bool cctlEnable) {
	for(const MemoryLayer& ml : mMemoryLayers) {
		if (!ml.mpPhysLayer)
			continue;

		if (ml.mbAutoRD4) {
			if (rightEnable)
				mpMemMan->ClearLayerMaskRange(ml.mpPhysLayer);
			else
				mpMemMan->SetLayerMaskRange(ml.mpPhysLayer, 0, 0);
		}

		if (ml.mbAutoRD5) {
			if (leftEnable)
				mpMemMan->ClearLayerMaskRange(ml.mpPhysLayer);
			else
				mpMemMan->SetLayerMaskRange(ml.mpPhysLayer, 0, 0);
		}

		if (ml.mbAutoCCTL) {
			if (cctlEnable)
				mpMemMan->ClearLayerMaskRange(ml.mpPhysLayer);
			else
				mpMemMan->SetLayerMaskRange(ml.mpPhysLayer, 0, 0);
		}
	}
}

void ATDeviceCustom::UpdateCartSense(bool leftActive) {
}

void ATDeviceCustom::InitIndicators(IATDeviceIndicatorManager *r) {
	mpIndicators = r;
}

void ATDeviceCustom::InitScheduling(ATScheduler *sch, ATScheduler *slowsch) {
	mpScheduler = sch;
}

void ATDeviceCustom::InitSIO(IATDeviceSIOManager *mgr) {
	mpSIOMgr = mgr;
}

IATDeviceSIO::CmdResponse ATDeviceCustom::OnSerialBeginCommand(const ATDeviceSIOCommand& cmd) {
	if (!cmd.mbStandardRate)
		return IATDeviceSIO::kCmdResponse_NotHandled;

	const SIOCommandTable *cmdTable = mpSIODeviceTable->mpDevices[cmd.mDevice];
	if (!cmdTable)
		return IATDeviceSIO::kCmdResponse_NotHandled;

	const SIOCommand *cmdEntry = cmdTable->mpCommands[cmd.mCommand];
	if (!cmdEntry)
		return IATDeviceSIO::kCmdResponse_Fail_NAK;

	if (cmdEntry->mpAutoTransferSegment) {
		mpActiveCommand = cmdEntry;

		mpSIOMgr->SendACK();

		if (cmdEntry->mbAutoTransferWrite) {
			mpSIOMgr->ReceiveData(0, cmdEntry->mAutoTransferLength, true);
			mpSIOMgr->SendComplete();
		} else {
			mpSIOMgr->SendComplete();
			mpSIOMgr->SendData(cmdEntry->mpAutoTransferSegment->mpData + cmdEntry->mAutoTransferOffset, cmdEntry->mAutoTransferLength, true);
		}

		mpSIOMgr->EndCommand();

		return IATDeviceSIO::kCmdResponse_Start;
	}

	if (!cmdEntry->mpScript)
		return IATDeviceSIO::kCmdResponse_Send_ACK_Complete;
	
	mpActiveCommand = cmdEntry;

	mVMThreadSIO.mThreadVariables[(int)ThreadVarIndex::Command] = cmd.mCommand;
	mVMThreadSIO.mThreadVariables[(int)ThreadVarIndex::Device] = cmd.mDevice;
	mVMThreadSIO.mThreadVariables[(int)ThreadVarIndex::Aux1] = cmd.mAUX[0];
	mVMThreadSIO.mThreadVariables[(int)ThreadVarIndex::Aux2] = cmd.mAUX[1];
	mVMThreadSIO.mThreadVariables[(int)ThreadVarIndex::Aux] = VDReadUnalignedLEU16(cmd.mAUX);
	mVMThreadSIO.mThreadVariables[(int)ThreadVarIndex::Timestamp] = ATSCHEDULER_GETTIME(mpScheduler);

	mSIO.mbDataFrameReceived = false;
	mSIO.mbValid = true;
	if (mVMThreadSIO.RunVoid(*cmdEntry->mpScript))
		mpSIOMgr->EndCommand();
	mSIO.mbValid = false;

	return IATDeviceSIO::kCmdResponse_Start;
}

void ATDeviceCustom::OnSerialAbortCommand() {
	mVMThreadSIO.Abort();
	mpActiveCommand = nullptr;
}

void ATDeviceCustom::OnSerialReceiveComplete(uint32 id, const void *data, uint32 len, bool checksumOK) {
	if (!mpActiveCommand)
		return;

	if (id == (uint32)SerialFenceId::AutoReceive) {
		memcpy(mpActiveCommand->mpAutoTransferSegment->mpData + mpActiveCommand->mAutoTransferOffset, data, len);
	} else if (id == (uint32)SerialFenceId::ScriptReceive) {
		mSIOFrameSegment.mpData = (uint8 *)data;
		mSIOFrameSegment.mSize = len;
		mSIO.mbValid = true;
		
		if (mVMThreadSIO.Resume())
			mpSIOMgr->EndCommand();

		mSIO.mbValid = false;
		mSIOFrameSegment.mpData = nullptr;
		mSIOFrameSegment.mSize = 0;
	}
}

void ATDeviceCustom::OnSerialFence(uint32 id) {
	if (id == (uint32)SerialFenceId::ScriptDelay) {
		mSIO.mbValid = true;

		if (mVMThreadSIO.Resume())
			mpSIOMgr->EndCommand();

		mSIO.mbValid = false;
	}
}

bool ATDeviceCustom::OnFileUpdated(const wchar_t *path) {
	if (CheckForTrackedChanges()) {
		ReloadConfig();

		if (mbInitedCustomDevice && mpIndicators)
			mpIndicators->SetStatusMessage(L"Custom device configuration reloaded.");
	}

	return true;
}

template<bool T_DebugOnly>
sint32 ATDeviceCustom::ReadControl(MemoryLayer& ml, uint32 addr) {
	const AddressBinding& binding = ml.mpReadBindings[addr - ml.mAddressBase];

	switch(binding.mAction) {
		case AddressAction::None:
			return -1;

		case AddressAction::ConstantData:
			return binding.mByteData;

		case AddressAction::Network:
			return 0xFF & SendNetCommand(addr, 0, T_DebugOnly ? NetworkCommand::DebugReadByte : NetworkCommand::ReadByte);

		case AddressAction::Script:
			mVMDomain.mSpecialVariables[(int)SpecialVarIndex::Address] = addr;
			mVMThread.mThreadVariables[(int)ThreadVarIndex::Timestamp] = ATSCHEDULER_GETTIME(mpScheduler);

			return 0xFF & mVMThread.RunInt(*mScriptFunctions[binding.mScriptFunctions[T_DebugOnly]]);

		case AddressAction::Variable:
			return 0xFF & mVMDomain.mGlobalVariables[binding.mVariableIndex];
	}

	return -1;
}

bool ATDeviceCustom::WriteControl(MemoryLayer& ml, uint32 addr, uint8 value) {
	const AddressBinding& binding = ml.mpWriteBindings[addr - ml.mAddressBase];

	switch(binding.mAction) {
		case AddressAction::None:
			return false;

		case AddressAction::Block:
			return true;

		case AddressAction::Network:
			SendNetCommand(addr, value, NetworkCommand::WriteByte);
			return true;

		case AddressAction::Script:
			mVMDomain.mSpecialVariables[(int)SpecialVarIndex::Address] = addr;
			mVMDomain.mSpecialVariables[(int)SpecialVarIndex::Value] = value;
			mVMThread.mThreadVariables[(int)ThreadVarIndex::Timestamp] = ATSCHEDULER_GETTIME(mpScheduler);

			mVMThread.RunVoid(*mScriptFunctions[binding.mScriptFunctions[0]]);
			return true;

		case AddressAction::Variable:
			mVMDomain.mGlobalVariables[binding.mVariableIndex] = value;
			return true;
	}

	return false;
}

bool ATDeviceCustom::PostNetCommand(uint32 address, sint32 value, NetworkCommand cmd) {
	if (!mpNetworkEngine)
		return false;

	uint8 cmdbuf[17];
	cmdbuf[0] = (uint8)cmd;
	VDWriteUnalignedLEU32(&cmdbuf[1], address);
	VDWriteUnalignedLEU32(&cmdbuf[5], value);
	VDWriteUnalignedLEU64(&cmdbuf[9], mpScheduler->GetTick64());

	for(;;) {
		if (mpNetworkEngine->Peek())
			PostNetError("Unexpected command received from server outside of event processing.");

		if (mpNetworkEngine->Send(cmdbuf, 17))
			break;

		if (!mpNetworkEngine->Restore())
			return false;
	}

	return true;
}

sint32 ATDeviceCustom::SendNetCommand(uint32 address, sint32 value, NetworkCommand cmd) {
	if (!mpNetworkEngine)
		return value;

	uint8 cmdbuf[16];

	for(;;) {
		if (!PostNetCommand(address, value, cmd))
			return value;

		for(;;) {
			if (!mpNetworkEngine->Recv(cmdbuf, 1))
				return 0;

			switch(cmdbuf[0]) {
				case (uint8)NetworkReply::None:
					break;

				case (uint8)NetworkReply::ReturnValue:
					mpNetworkEngine->Recv(cmdbuf, 4);
					return VDReadUnalignedLEU32(&cmdbuf[0]);

				case (uint8)NetworkReply::EnableMemoryLayer:
					mpNetworkEngine->Recv(cmdbuf, 2);
					if (cmdbuf[0] < mMemoryLayers.size()) {
						auto *ml = &mMemoryLayers[cmdbuf[0]];

						if (ml->mpPhysLayer) {
							uint32 modes = 0;

							if (cmdbuf[1] & 0x01)
								modes |= kATMemoryAccessMode_W;

							if (cmdbuf[1] & 0x02)
								modes |= kATMemoryAccessMode_AR;

							mpMemMan->SetLayerModes(ml->mpPhysLayer, (ATMemoryAccessMode)modes);

							if (ml->mbAutoRD5) {
								bool rd5Active = (modes != 0);

								if (ml->mbRD5Active != rd5Active) {
									ml->mbRD5Active = rd5Active;
									mpCartPort->OnLeftWindowChanged(mCartId, IsLeftCartActive());
								}
							}
						}
					} else
						PostNetError("EnableMemoryLayer: Invalid memory layer index");
					break;

				case (uint8)NetworkReply::SetMemoryLayerOffset:
					mpNetworkEngine->Recv(cmdbuf, 5);
					if (cmdbuf[0] < mMemoryLayers.size()) {
						auto *ml = &mMemoryLayers[cmdbuf[0]];

						if (ml->mpPhysLayer && ml->mpSegment) {
							uint32 offset = VDReadUnalignedLEU32(&cmdbuf[1]);

							if (!(offset & 0xFF) && offset <= ml->mMaxOffset)
								mpMemMan->SetLayerMemory(ml->mpPhysLayer, ml->mpSegment->mpData + offset);
						} else
							PostNetError("SetMemoryLayerOffset: Invalid memory layer offset");
					} else
						PostNetError("SetMemoryLayerOffset: Invalid memory layer index");
					break;

				case (uint8)NetworkReply::SetMemoryLayerSegmentOffset:
					mpNetworkEngine->Recv(cmdbuf, 6);
					if (cmdbuf[0] < mMemoryLayers.size() && cmdbuf[1] < mSegments.size()) {
						auto& ml = mMemoryLayers[cmdbuf[0]];
						auto& seg = mSegments[cmdbuf[1]];

						uint32 offset = VDReadUnalignedLEU32(&cmdbuf[2]);
						if (ml.mpPhysLayer && !(offset & 0xFF) && offset < seg.mSize && seg.mSize - offset >= ml.mSize) {
							ml.mpSegment = &seg;
							ml.mMaxOffset = seg.mSize - ml.mSize;

							mpMemMan->SetLayerMemory(ml.mpPhysLayer, seg.mpData + offset);
						} else
							PostNetError("SetMemoryLayerSegmentOffset: Invalid memory layer range");
					} else
						PostNetError("SetMemoryLayerSegmentOffset: Invalid memory layer index");
					break;

				case (uint8)NetworkReply::SetMemoryLayerReadOnly:
					mpNetworkEngine->Recv(cmdbuf, 2);
					if (cmdbuf[0] < mMemoryLayers.size()) {
						auto *ml = &mMemoryLayers[cmdbuf[0]];

						if (ml->mpPhysLayer)
							mpMemMan->SetLayerReadOnly(ml->mpPhysLayer, cmdbuf[1] != 0);
					} else
						PostNetError("SetMemoryLayerReadOnly: Invalid memory layer index");
					break;

				case (uint8)NetworkReply::ReadSegmentMemory:
					mpNetworkEngine->Recv(cmdbuf, 9);
					if (cmdbuf[0] < mSegments.size()) {
						auto& seg = mSegments[cmdbuf[0]];
						uint32 offset = VDReadUnalignedLEU32(&cmdbuf[1]);
						uint32 len = VDReadUnalignedLEU32(&cmdbuf[5]);

						if (offset <= seg.mSize && seg.mSize - offset >= len) {
							mpNetworkEngine->Send(seg.mpData + offset, len);
						} else
							PostNetError("ReadSegmentMemory: Invalid segment range");
					} else
						PostNetError("ReadSegmentMemory: Invalid segment index");
					break;

				case (uint8)NetworkReply::WriteSegmentMemory:
					mpNetworkEngine->Recv(cmdbuf, 9);
					if (cmdbuf[0] < mSegments.size()) {
						auto& seg = mSegments[cmdbuf[0]];
						uint32 offset = VDReadUnalignedLEU32(&cmdbuf[1]);
						uint32 len = VDReadUnalignedLEU32(&cmdbuf[5]);

						if (offset <= seg.mSize && seg.mSize - offset >= len) {
							mpNetworkEngine->Recv(seg.mpData + offset, len);
						} else
							PostNetError("WriteSegmentMemory: Invalid segment range");
					} else
						PostNetError("WriteSegmentMemory: Invalid segment index");
					break;

				case (uint8)NetworkReply::CopySegmentMemory:
					mpNetworkEngine->Recv(cmdbuf, 14);
					if (cmdbuf[0] < mSegments.size() && cmdbuf[5] < mSegments.size()) {
						auto& dstSeg = mSegments[cmdbuf[0]];
						uint32 dstOffset = VDReadUnalignedLEU32(&cmdbuf[1]);
						auto& srcSeg = mSegments[cmdbuf[5]];
						uint32 srcOffset = VDReadUnalignedLEU32(&cmdbuf[6]);
						uint32 len = VDReadUnalignedLEU32(&cmdbuf[10]);

						if (dstOffset <= dstSeg.mSize && dstSeg.mSize - dstOffset >= len
							&& srcOffset <= srcSeg.mSize && srcSeg.mSize - srcOffset >= len)
						{
							memmove(dstSeg.mpData + dstOffset, srcSeg.mpData + srcOffset, len);
						} else {
							PostNetError("CopySegmentMemory: Invalid segment ranges");
						}
					} else
						PostNetError("CopySegmentMemory: Invalid segment index");
					break;

				default:
					PostNetError("Invalid command received");
			}
		}
	}
}

void ATDeviceCustom::PostNetError(const char *msg) {
	if (mpNetworkEngine) {
		uint32 len = (uint32)strlen(msg);

		// send the command frame directly as to avoid a loop
		uint8 cmdbuf[17];
		cmdbuf[0] = (uint8)NetworkCommand::Error;
		VDWriteUnalignedLEU32(&cmdbuf[1], 0);
		VDWriteUnalignedLEU32(&cmdbuf[5], len);
		VDWriteUnalignedLEU64(&cmdbuf[9], mpScheduler->GetTick64());
		mpNetworkEngine->Send(cmdbuf, 17);

		mpNetworkEngine->Send(msg, len);
	}

	if (mpIndicators) {
		VDStringW s;
		s.sprintf(L"Communication error with custom device server: %hs", msg);
		mpIndicators->ReportError(s.c_str());
	}
}

void ATDeviceCustom::ShutdownCustomDevice() {
	auto& sem = *g_sim.GetEventManager();

	if (mEventBindingVBLANK) {
		sem.RemoveEventCallback(mEventBindingVBLANK);
		mEventBindingVBLANK = 0;
	}

	mTrackedFileWatcher.Shutdown();

	mbInitedCustomDevice = false;
	mpActiveCommand = nullptr;

	mDeviceName.clear();

	if (mpNetworkEngine) {
		mpNetworkEngine->Shutdown();
		mpNetworkEngine = nullptr;
	}

	if (mbInitedCart) {
		mbInitedCart = false;

		mpCartPort->RemoveCartridge(mCartId, this);
	}

	if (mbInitedSIO) {
		mbInitedSIO = false;

		mpSIOMgr->RemoveDevice(this);
	}

	for(MemoryLayer& ml : mMemoryLayers) {
		if (ml.mpPhysLayer)
			mpMemMan->DeleteLayerPtr(&ml.mpPhysLayer);
	}

	for(auto& cport : mControllerPorts)
		cport.Shutdown();

	mMemoryLayers = {};
	mSegments = {};
	mpSIODeviceTable = nullptr;

	mpScriptEventColdReset = nullptr;
	mpScriptEventWarmReset = nullptr;
	mpScriptEventInit = nullptr;

	mConfigAllocator.Clear();
	mVMDomain.Clear();
}

void ATDeviceCustom::ReinitSegmentData(bool clearNonVolatile) {
	for(const Segment& seg : mSegments) {
		if (!clearNonVolatile && seg.mbNonVolatile)
			continue;

		uint8 *dst = seg.mpData;
		const uint32 len = seg.mSize;

		if (!dst)
			continue;
		
		const uint8 *pat = seg.mpInitData;
		const uint32 patlen = seg.mInitSize;

		if (seg.mInitSize == 1)
			if (pat)
				memset(dst, pat[0], len);
			else
				memset(dst, 0, len);
		else if (seg.mInitSize == 2) {
			VDMemset16(dst, VDReadUnalignedU16(pat), len >> 1);
			memcpy(dst + (len & ~1), pat, len & 1);
		} else if (seg.mInitSize == 4) {
			VDMemset32(dst, VDReadUnalignedU32(pat), len >> 1);
			memcpy(dst + (len & ~3), pat, len & 3);
		} else if (pat) {
			// we are guaranteed that the segment is at least as big as the pattern,
			// as this is enforced by the parser
			memcpy(dst, pat, patlen);

			const uint8 *src = dst;
			dst += patlen;

			// do a deliberately self-referencing ascending copy
			for(uint32 i = patlen; i < len; ++i)
				*dst++ = *src++;
		}
	}
}

void ATDeviceCustom::ReloadConfig() {
	ShutdownCustomDevice();
	ClearFileTracking();

	try {
		{
			vdrefptr<ATVFSFileView> view;
			OpenViewWithTracking(mConfigPath.c_str(), ~view);

			mResourceBasePath.assign(mConfigPath.c_str(), ATVFSSplitPathFile(mConfigPath.c_str()));

			IVDRandomAccessStream& stream = view->GetStream();
			sint64 len = stream.Length();

			if ((uint64)len >= kMaxRawConfigSize)
				throw MyError("Custom device description is too large: %llu bytes.", (unsigned long long)len);

			vdblock<char> buf;
			buf.resize((size_t)len);
			stream.Read(buf.data(), len);

			VDJSONDocument doc(true);
			VDJSONReader reader;

			reader.SetMemberNameFilter(
				[](const wchar_t *name) {
					return (name[0] != '/' || name[1] != '/');
				}
			);

			if (!reader.Parse(buf.data(), buf.size(), doc)) {
				int line, offset;
				reader.GetErrorLocation(line, offset);

				throw MyError("Custom device description failed to load: invalid JSON syntax at line %u", line);
			}

			ProcessDesc(doc);
		}

		bool needCart = false;

		for(MemoryLayer& ml : mMemoryLayers) {
			if (ml.mbAutoRD4 || ml.mbAutoRD5 || ml.mbAutoCCTL) {
				needCart = true;
				break;
			}
		}

		if (needCart) {
			mpCartPort->AddCartridge(this, kATCartridgePriority_Default, mCartId);
			mpCartPort->OnLeftWindowChanged(mCartId, IsLeftCartActive());

			mbInitedCart = true;
		}

		if (mpSIODeviceTable) {
			mpSIOMgr->AddDevice(this);
			mbInitedSIO = true;
		}

		ReinitSegmentData(true);

		if (mNetworkPort) {
			mpNetworkEngine = ATCreateDeviceCustomNetworkEngine(mNetworkPort);
			if (!mpNetworkEngine)
				throw MyError("Unable to set up local networking for custom device.");
		}

		if (mpScriptEventVBLANK) {
			mEventBindingVBLANK = g_sim.GetEventManager()->AddEventCallback(kATSimEvent_VBLANK,
				[this] { mVMThread.RunVoid(*mpScriptEventVBLANK); });
		}

		mbInitedCustomDevice = true;
		mLastError.clear();

		if (mpScriptEventInit)
			mVMThread.RunVoid(*mpScriptEventInit);


	} catch(const MyError& e) {
		ShutdownCustomDevice();

		mLastError = VDTextAToW(e.gets());

		if (mpIndicators)
			mpIndicators->ReportError(mLastError.c_str());
	}

	// We need to set up tracking even if the load failed, so we can try again when the file is fixed
	if (!mTrackedFiles.empty() && mbHotReload) {
		try {
			mTrackedFileWatcher.InitDir(VDFileSplitPathLeft(mTrackedFiles.begin()->first).c_str(), false, this);
		} catch(const MyError&) {
		}
	}
}

class ATDeviceCustom::MemberParser {
public:
	MemberParser(const VDJSONValueRef& object);

	VDJSONValueRef Required(const char *name);
	VDJSONValueRef Required(uint32 nameToken);
	VDJSONValueRef Optional(const char *name);
	VDJSONValueRef Optional(uint32 nameToken);

	uint32 RequiredUint32(const char *name);
	const wchar_t *RequiredString(const char *name);

	void AssertNoUnused();

	VDJSONValueRef mRootObject;
	const VDJSONNameTable *mpNameTable;
	vdhashmap<uint32, VDJSONValueRef> mMembers;
};

ATDeviceCustom::MemberParser::MemberParser(const VDJSONValueRef& object)
	: mRootObject(object)
{
	if (!object.IsObject())
		throw ATInvalidCustomDeviceDescException(object, "Expected object");

	mpNameTable = &object.GetDocument()->mNameTable;

	for(const VDJSONMemberRef& memberRef : object.AsObject()) {
		mMembers[memberRef.GetNameToken()] = memberRef.GetValue();
	}
}

VDJSONValueRef ATDeviceCustom::MemberParser::Required(const char *name) {
	return Required(mpNameTable->GetToken(name).mToken);
}

VDJSONValueRef ATDeviceCustom::MemberParser::Required(uint32 nameToken) {
	auto val = Optional(nameToken);

	if (!val.IsValid())
		throw ATInvalidCustomDeviceDescExceptionF(mRootObject, "Required member '%ls' not found", mpNameTable->GetName(nameToken));

	return val;
}

VDJSONValueRef ATDeviceCustom::MemberParser::Optional(const char *name) {
	return Optional(mpNameTable->GetToken(name).mToken);
}

VDJSONValueRef ATDeviceCustom::MemberParser::Optional(uint32 nameToken) {
	if (!nameToken)
		return {};

	auto it = mMembers.find(nameToken);

	if (it == mMembers.end())
		return {};

	auto val = it->second;

	mMembers.erase(it);

	return val;
}

uint32 ATDeviceCustom::MemberParser::RequiredUint32(const char *name) {
	auto val = Required(name);

	return ParseRequiredUint32(val);
}

const wchar_t *ATDeviceCustom::MemberParser::RequiredString(const char *name) {
	auto val = Required(name);

	return ParseRequiredString(val);
}

void ATDeviceCustom::MemberParser::AssertNoUnused() {
	if (!mMembers.empty()) 
		throw ATInvalidCustomDeviceDescExceptionF(mMembers.begin()->second, "Unexpected member '%ls'", mpNameTable->GetName(mMembers.begin()->first));
}

void ATDeviceCustom::ProcessDesc(const VDJSONDocument& doc) {
	ATCustomDeviceCompiler compiler(mVMDomain);
	compiler.DefineSpecialVariable("address");
	compiler.DefineSpecialVariable("value");
	compiler.DefineThreadVariable("timestamp");
	compiler.DefineThreadVariable("device");
	compiler.DefineThreadVariable("command");
	compiler.DefineThreadVariable("aux1");
	compiler.DefineThreadVariable("aux2");
	compiler.DefineThreadVariable("aux");
	compiler.DefineSpecialObjectVariable("network", &mNetwork);
	compiler.DefineSpecialObjectVariable("sio_frame", &mSIOFrameSegment);
	compiler.DefineSpecialObjectVariable("sio", &mSIO);
	compiler.DefineSpecialObjectVariable("clock", &mClock);

	mpCompiler = &compiler;

	const VDJSONValueRef& rootNode = doc.Root();

	if (!rootNode.IsObject())
		throw ATInvalidCustomDeviceDescException(rootNode, "Top-level item is not an object.");

	MemberParser rootMembers(rootNode);

	const VDJSONValueRef& nameNode = rootMembers.Optional("name");
	if (nameNode.IsValid())
		mDeviceName = ParseRequiredString(nameNode);

	mNetworkPort = 0;
	const VDJSONValueRef& networkNode = rootMembers.Optional("network");
	if (networkNode.IsValid()) {
		MemberParser networkMembers(networkNode);
		const VDJSONValueRef& portNode = networkMembers.Required("port");
		uint32 port = ParseRequiredUint32(portNode);

		if (port < 1024 || port >= 49151)
			throw ATInvalidCustomDeviceDescException(portNode, "Invalid network port (not in 1024-49150).");

		mNetworkPort = port;

		networkMembers.AssertNoUnused();
	}

	const VDJSONValueRef& variablesNode = rootMembers.Optional("variables");
	if (variablesNode.IsValid()) {
		if (!variablesNode.IsObject())
			throw ATInvalidCustomDeviceDescException(variablesNode, "Variables item must be an object.");

		for(const VDJSONMemberRef& variableRef : variablesNode.AsObject()) {
			VDStringSpanW type{ParseRequiredString(variableRef.GetValue())};

			if (type == L"int") {
				mpCompiler->DefineIntegerVariable(VDTextWToU8(VDStringSpanW(variableRef.GetName())).c_str());
			} else
				throw ATInvalidCustomDeviceDescException(variablesNode, "Invalid type for variable.");
		}
	}

	// instantiate segments and segment script objects
	const VDJSONValueRef& segmentsNode = rootMembers.Optional("segments");
	if (segmentsNode.IsValid())
		ProcessSegments(segmentsNode);

	// precreate memory layers to establish script objects
	const VDJSONValueRef& memoryLayersNode = rootMembers.Optional("memory_layers");
	if (memoryLayersNode.IsValid())
		ProcessMemoryLayers(memoryLayersNode);

	if (const VDJSONValueRef& controllerPortsNode = rootMembers.Optional("controller_ports"); controllerPortsNode.IsValid())
		ProcessControllerPorts(controllerPortsNode);

	// process script includes
	if (const VDJSONValueRef& scriptIncludesNode = rootMembers.Optional("script_includes"); scriptIncludesNode.IsValid()) {
		if (!scriptIncludesNode.IsArray())
			throw ATInvalidCustomDeviceDescException(scriptIncludesNode, "script_includes must be an array.");

		for(const VDJSONValueRef& scriptIncludeNode : scriptIncludesNode.AsArray()) {
			vdrefptr<ATVFSFileView> sourceView;		
			LoadDependency(scriptIncludeNode, ~sourceView);

			IVDRandomAccessStream& stream = sourceView->GetStream();
			sint64 len = stream.Length();

			if ((uint64)len > 16 * 1024 * 1024)
				throw ATInvalidCustomDeviceDescException(scriptIncludesNode, "Script include is too large");

			vdblock<char> incdata((uint32)len);
			stream.Read(incdata.data(), incdata.size());

			if (!mpCompiler->CompileFile(incdata.data(), incdata.size())) {
				VDStringA s;

				auto [line, column] = mpCompiler->GetErrorLinePos();
				s.sprintf("%ls(%u,%u): %s", sourceView->GetFileName(), line, column, mpCompiler->GetError());
				throw ATInvalidCustomDeviceDescException(scriptIncludesNode, s.c_str());
			}
		}
	}

	// finish memory layer processing now that all objects and script includes have been processed
	if (memoryLayersNode.IsValid())
		ProcessMemoryLayers2(memoryLayersNode);

	if (const VDJSONValueRef& sioDevicesNode = rootMembers.Optional("sio_devices"); sioDevicesNode.IsValid())
		ProcessSIODevices(sioDevicesNode);

	mVMThread.Init(mVMDomain);
	mVMThreadSIO.Init(mVMDomain);

	const VDJSONValueRef& scriptEventsNode = rootMembers.Optional("script_events");
	if (scriptEventsNode.IsValid()) {
		MemberParser scriptEvents(scriptEventsNode);

		mpScriptEventInit = ParseScriptOpt(kATCDVMTypeVoid, scriptEvents.Optional("init"));
		mpScriptEventColdReset = ParseScriptOpt(kATCDVMTypeVoid, scriptEvents.Optional("cold_reset"));
		mpScriptEventWarmReset = ParseScriptOpt(kATCDVMTypeVoid, scriptEvents.Optional("warm_reset"));;
		mpScriptEventVBLANK = ParseScriptOpt(kATCDVMTypeVoid, scriptEvents.Optional("vblank"));;
	}

	rootMembers.AssertNoUnused();
}

void ATDeviceCustom::ProcessSegments(VDJSONValueRef segmentsNode) {
	if (!segmentsNode.IsObject())
		throw ATInvalidCustomDeviceDescException(segmentsNode, "Segments item is not an object.");

	const size_t numSegs = segmentsNode.GetMemberCount();
	Segment *segs = (Segment *)mConfigAllocator.Allocate(sizeof(Segment) * numSegs, alignof(Segment));

	for(size_t i=0; i<numSegs; ++i)
		new(&segs[i]) Segment();

	mSegments = vdvector_view(segs, numSegs);

	for(const VDJSONMemberRef& segmentRef : segmentsNode.AsObject())
		ProcessSegment(*segs++, segmentRef);
}

void ATDeviceCustom::ProcessSegment(Segment& seg, VDJSONMemberRef memberRef) {
	const VDJSONValueRef& objectNode = memberRef.GetValue();
	if (!objectNode.IsObject())
		throw ATInvalidCustomDeviceDescException(objectNode, "Segment item is not an object.");

	MemberParser segMembers(objectNode);

	const uint32 size = segMembers.RequiredUint32("size");
	if (size > kMaxTotalSegmentData)
		throw ATUnsupportedCustomDeviceDescException(objectNode, "Total segment size is too large.");

	seg.mId = memberRef.GetNameToken();
	seg.mpData = (uint8 *)mConfigAllocator.Allocate(size, 4);
	if (!seg.mpData)
		throw MyMemoryError();

	seg.mSize = size;
	seg.mbNonVolatile = false;
	seg.mbMappable = true;

	if (const VDJSONValueRef& sourceNode = segMembers.Optional("source"); sourceNode.IsValid()) {
		vdrefptr<ATVFSFileView> sourceView;		
		LoadDependency(sourceNode, ~sourceView);

		const VDJSONValueRef& offsetNode = objectNode["source_offset"];
		uint32 offset = offsetNode.IsValid() ? ParseRequiredUint32(offsetNode) : 0;

		IVDRandomAccessStream& stream = sourceView->GetStream();
		seg.mpInitData = (uint8 *)mConfigAllocator.Allocate(size, 4);
		seg.mInitSize = size;

		if (offset)
			stream.Seek(offset);

		stream.Read(seg.mpInitData, seg.mInitSize);
	} else if (const VDJSONValueRef& initPatternNode = segMembers.Optional("init_pattern"); initPatternNode.IsValid()) {
		auto pattern = ParseBlob(initPatternNode);

		if (pattern.size() > seg.mSize)
			throw ATInvalidCustomDeviceDescException(initPatternNode, "Init pattern is larger than the segment size.");

		seg.mpInitData = pattern.data();
		seg.mInitSize = (uint32)pattern.size();
	} else {
		seg.mInitSize = 1;
	}

	if (const VDJSONValueRef& persistenceNode = segMembers.Optional("persistence"); persistenceNode.IsValid()) {
		const VDStringSpanW mode(ParseRequiredString(persistenceNode));

		if (mode == L"nonvolatile") {
			seg.mbNonVolatile = true;
		} else if (mode != L"volatile")
			throw ATInvalidCustomDeviceDescException(persistenceNode, "Unknown segment persistence mode.");
	}

	segMembers.AssertNoUnused();

	if (!mpCompiler->DefineObjectVariable(VDTextWToU8(VDStringSpanW(memberRef.GetName())).c_str(), &seg, Segment::kVMObjectClass))
		throw ATInvalidCustomDeviceDescException(objectNode, mpCompiler->GetError());
}

void ATDeviceCustom::ProcessMemoryLayers(VDJSONValueRef memoryLayersNode) {
	if (!memoryLayersNode.IsObject())
		throw ATInvalidCustomDeviceDescException(memoryLayersNode, "Memory layers item is not an object.");

	const size_t numLayers = memoryLayersNode.GetMemberCount();
	MemoryLayer *layers = (MemoryLayer *)mConfigAllocator.Allocate(sizeof(MemoryLayer) * numLayers, alignof(MemoryLayer));

	for(size_t i=0; i<numLayers; ++i)
		new(&layers[i]) MemoryLayer();

	mMemoryLayers = vdvector_view(layers, numLayers);

	// predefine the memory layer variables before we process any scripts
	for(const VDJSONMemberRef& memoryLayerRef : memoryLayersNode.AsObject()) {
		if (!mpCompiler->DefineObjectVariable(VDTextWToU8(VDStringSpanW(memoryLayerRef.GetName())).c_str(), layers++, MemoryLayer::kVMObjectClass))
			throw ATInvalidCustomDeviceDescException(memoryLayerRef.GetValue(), mpCompiler->GetError());
	}
}

void ATDeviceCustom::ProcessMemoryLayers2(VDJSONValueRef memoryLayersNode) {
	MemoryLayer *layers =  mMemoryLayers.data();

	for(const VDJSONMemberRef& memoryLayerRef : memoryLayersNode.AsObject())
		ProcessMemoryLayer(*layers++, memoryLayerRef);
}

void ATDeviceCustom::ProcessMemoryLayer(MemoryLayer& ml, VDJSONMemberRef memberRef) {
	const VDJSONValueRef& objectNode = memberRef.GetValue();
	if (!objectNode.IsObject())
		throw ATInvalidCustomDeviceDescException(objectNode, "Memory layer item is not an object.");

	MemberParser mlMembers(objectNode);

	const uint32 address = mlMembers.RequiredUint32("address");
	const uint32 size = mlMembers.RequiredUint32("size");

	if (address >= 0x10000 || (address & 0xFF))
		throw ATInvalidCustomDeviceDescException(objectNode, "Memory layer address is invalid.");

	if (size == 0 || 0x10000 - address < size || (size & 0xFF))
		throw ATInvalidCustomDeviceDescException(objectNode, "Memory layer size is invalid.");

	ml.mpParent = this;
	ml.mId = memberRef.GetNameToken();
	ml.mAddressBase = address;
	ml.mSize = size;

	// see if we have a fixed mapping or control mapping
	VDJSONValueRef fixedMappingNode = mlMembers.Optional("fixed_mapping");

	if (!fixedMappingNode.IsValid())
		fixedMappingNode = mlMembers.Optional("segment");

	if (fixedMappingNode.IsValid()) {
		MemberParser segmentMembers(fixedMappingNode);
		const VDJSONValueRef& sourceNode = segmentMembers.Required("source");
		const wchar_t *sourceName = ParseRequiredString(sourceNode);

		uint32 id = fixedMappingNode.GetDocument()->mNameTable.GetToken(sourceName).mToken;
		Segment *srcSegment = nullptr;

		for(Segment& seg : mSegments) {
			if (seg.mId == id) {
				srcSegment = &seg;
				break;
			}
		}

		if (!srcSegment)
			throw ATInvalidCustomDeviceDescException(sourceNode, "The specified source segment could not be found.");

		// parse byte offset within segment
		const VDJSONValueRef& offsetNode = segmentMembers.Optional("offset");
		uint32 offset = offsetNode.IsValid() ? ParseRequiredUint32(offsetNode) : 0;

		// validate that it is page aligned and within the source segment
		if (offset & 0xFF)
			throw ATInvalidCustomDeviceDescException(offsetNode, "Source segment offset is invalid.");

		if (offset >= srcSegment->mSize || srcSegment->mSize - offset < size)
			throw ATInvalidCustomDeviceDescException(offsetNode, "Reference extends outside of source segment.");

		ml.mpSegment = srcSegment;
		ml.mMaxOffset = srcSegment->mSize - size;

		// parse mode
		const VDJSONValueRef& modeNode = segmentMembers.Required("mode");
		const wchar_t *mode = ParseRequiredString(modeNode);
		ATMemoryAccessMode accessMode;
		bool isReadOnly = false;

		if (!wcscmp(mode, L"r")) {
			accessMode = kATMemoryAccessMode_AR;
		} else if (!wcscmp(mode, L"ro")) {
			accessMode = kATMemoryAccessMode_ARW;
			isReadOnly = true;
		} else if (!wcscmp(mode, L"w")) {
			accessMode = kATMemoryAccessMode_W;
		} else if (!wcscmp(mode, L"rw")) {
			accessMode = kATMemoryAccessMode_ARW;
		} else
			throw ATInvalidCustomDeviceDescException(offsetNode, "Invalid access mode.");

		ml.mpPhysLayer = mpMemMan->CreateLayer(kATMemoryPri_Cartridge1, srcSegment->mpData + offset, address >> 8, size >> 8, isReadOnly); 
		mpMemMan->SetLayerModes(ml.mpPhysLayer, accessMode);

		segmentMembers.AssertNoUnused();
	} else if (const VDJSONValueRef& controlNode = mlMembers.Optional("control"); controlNode.IsValid()) {
		// parse address actions
		if (!controlNode.IsArray())
			throw ATInvalidCustomDeviceDescException(controlNode, "Control item was not an object.");

		ml.mpReadBindings = mConfigAllocator.AllocateArray<AddressBinding>(size);
		ml.mpWriteBindings = mConfigAllocator.AllocateArray<AddressBinding>(size);

		uint32 memoryLayerMode = 0;

		for(const VDJSONValueRef bindingNode : controlNode.AsArray()) {
			if (!bindingNode.IsObject())
				throw ATInvalidCustomDeviceDescException(bindingNode, "Binding item  was not an object.");

			MemberParser bindingMembers(bindingNode);

			// parse address attribute and validate it
			const VDJSONValueRef& bindingAddressNode = bindingMembers.Required("address");
			uint32 actionAddress = ParseRequiredUint32(bindingAddressNode);

			// parse optional size attribute
			const VDJSONValueRef& sizeNode = bindingMembers.Optional("size");
			uint32 actionSize = sizeNode.IsValid() ? ParseRequiredUint32(sizeNode) : 1;

			// validate that address action is within layer
			if (actionAddress < address || actionAddress >= address + size || (address + size) - actionAddress < actionSize)
				throw ATInvalidCustomDeviceDescException(bindingAddressNode, "Binding address is outside of the memory layer.");

			// parse mode
			const VDJSONValueRef& bindingActionModeNode = bindingMembers.Required("mode");
			const wchar_t *actionMode = ParseRequiredString(bindingActionModeNode);
			bool actionRead = false;
			bool actionWrite = false;

			if (!wcscmp(actionMode, L"r"))
				actionRead = true;
			else if (!wcscmp(actionMode, L"w"))
				actionWrite = true;
			else if (!wcscmp(actionMode, L"rw"))
				actionRead = actionWrite = true;
			else
				throw ATInvalidCustomDeviceDescException(bindingActionModeNode, "Invalid binding mode.");

			// validate that actions are not already assigned
			vdvector_view<AddressBinding> actionReadBindings;
			vdvector_view<AddressBinding> actionWriteBindings;

			if (actionRead) {
				actionReadBindings = { &ml.mpReadBindings[actionAddress - address], actionSize };

				for(const auto& rb : actionReadBindings) {
					if (rb.mAction != AddressAction::None)
						throw ATInvalidCustomDeviceDescException(bindingNode, "Address conflict between two read bindings in the same layer.");
				}

				memoryLayerMode |= kATMemoryAccessMode_AR;
			}

			if (actionWrite) {
				actionWriteBindings = { &ml.mpWriteBindings[actionAddress - address], actionSize };

				for(const auto& wb : actionWriteBindings) {
					if (wb.mAction != AddressAction::None)
						throw ATInvalidCustomDeviceDescException(bindingNode, "Address conflict between two write bindings in the same layer.");
				}

				memoryLayerMode |= kATMemoryAccessMode_W;
			}

			if (const VDJSONValueRef& dataNode = bindingMembers.Optional("data"); dataNode.IsValid()) {
				const vdvector_view<uint8> pattern = ParseBlob(dataNode);

				if (!actionRead || actionWrite)
					throw ATInvalidCustomDeviceDescException(dataNode, "Data bindings can only be read-only.");

				if (pattern.size() == 1) {
					for(auto& rb : actionReadBindings) {
						rb.mAction = AddressAction::ConstantData;
						rb.mByteData = pattern[0];
					}
				} else {
					if (pattern.size() != actionSize)
						throw ATInvalidCustomDeviceDescException(dataNode, "Data must either be a single byte or the same size as the memory layer.");

					for(uint32 i = 0; i < actionSize; ++i) {
						auto& arb = actionReadBindings[i];

						arb.mAction = AddressAction::ConstantData;
						arb.mByteData = pattern[i];
					}
				}
			} else if (const VDJSONValueRef& actionNode = bindingMembers.Optional("action"); actionNode.IsValid()) {
				const wchar_t *actionStr = actionNode.AsString();

				if (!wcscmp(actionStr, L"block")) {
					if (actionRead || !actionWrite)
						throw ATInvalidCustomDeviceDescException(actionNode, "Block bindings can only be write-only.");

					for(auto& wb : actionWriteBindings)
						wb.mAction = AddressAction::Block;
				} else if (!wcscmp(actionStr, L"network")) {
					if (mNetworkPort == 0)
						throw ATInvalidCustomDeviceDescException(actionNode, "Cannot use a network binding with no network connection set up.");

					for(auto& rb : actionReadBindings)
						rb.mAction = AddressAction::Network;

					for(auto& wb : actionWriteBindings)
						wb.mAction = AddressAction::Network;
				} else
					throw ATInvalidCustomDeviceDescException(actionNode, "Unknown action type.");
			} else if (const VDJSONValueRef& variableNode = bindingMembers.Optional("variable"); variableNode.IsValid()) {
				const wchar_t *varNameStr = ParseRequiredString(variableNode);

				const ATCDVMTypeInfo *varTypeInfo = mpCompiler->GetVariable(VDTextWToU8(VDStringSpanW(varNameStr)).c_str());

				if (!varTypeInfo)
					throw ATInvalidCustomDeviceDescException(variableNode, "Variable not defined.");

				if (varTypeInfo->mClass != ATCDVMTypeClass::IntLValue)
					throw ATInvalidCustomDeviceDescException(variableNode, "Variable must be of integer type for address binding.");

				for(auto& rb : actionReadBindings) {
					rb.mAction = AddressAction::Variable;
					rb.mVariableIndex = varTypeInfo->mIndex;
				}

				for(auto& wb : actionWriteBindings) {
					wb.mAction = AddressAction::Variable;
					wb.mVariableIndex = varTypeInfo->mIndex;
				}
			} else if (const VDJSONValueRef& scriptNode = bindingMembers.Optional("script"); scriptNode.IsValid()) {
				const ATCDVMFunction *func = ParseScript(actionRead ? kATCDVMTypeInt : kATCDVMTypeVoid, scriptNode);

				const ATCDVMFunction *debugFunc = nullptr;
				if (const VDJSONValueRef& debugScriptNode = bindingMembers.Optional("debug_script"); debugScriptNode.IsValid())
					debugFunc = ParseScript(kATCDVMTypeInt, debugScriptNode);

				uint16 funcCode = (uint16)mScriptFunctions.size();
				mScriptFunctions.push_back(func);

				uint16 debugFuncCode = funcCode;
				if (debugFunc) {
					debugFuncCode = (uint16)mScriptFunctions.size();
					mScriptFunctions.push_back(debugFunc);
				}


				for(auto& rb : actionReadBindings) {
					rb.mAction = AddressAction::Script;
					rb.mScriptFunctions[0] = funcCode;
					rb.mScriptFunctions[1] = debugFuncCode;
				}

				for(auto& wb : actionWriteBindings) {
					wb.mAction = AddressAction::Script;
					wb.mScriptFunctions[0] = funcCode;
				}
			} else if (const VDJSONValueRef& copyNode = bindingMembers.Optional("copy_from"); copyNode.IsValid()) {
				const uint32 srcAddr = ParseRequiredUint32(copyNode);

				if (srcAddr < address || (srcAddr - address) >= size || size - (srcAddr - address) < actionSize)
					throw ATInvalidCustomDeviceDescException(bindingNode, "Binding copy source is outside of memory layer.");

				const auto *rsrc = &ml.mpReadBindings[srcAddr - address];
				for(auto& rb : actionReadBindings)
					rb = *rsrc++;

				const auto *wsrc = &ml.mpWriteBindings[srcAddr - address];
				for(auto& wb : actionWriteBindings)
					wb = *wsrc++;
			} else
				throw ATInvalidCustomDeviceDescException(bindingNode, "No binding type specified.");

			bindingMembers.AssertNoUnused();
		}

		if (!memoryLayerMode)
			throw ATInvalidCustomDeviceDescException(controlNode, "No address bindings were specified.");

		// create physical memory layer
		ATMemoryHandlerTable handlers {};
		handlers.mbPassAnticReads = true;
		handlers.mbPassReads = true;
		handlers.mbPassWrites = true;
		handlers.mpThis = &ml;

		handlers.mpDebugReadHandler = [](void *thisptr, uint32 addr) -> sint32 {
			MemoryLayer& ml = *(MemoryLayer *)thisptr;

			return ml.mpParent->ReadControl<true>(ml, addr);
		};

		handlers.mpReadHandler = [](void *thisptr, uint32 addr) -> sint32 {
			MemoryLayer& ml = *(MemoryLayer *)thisptr;

			return ml.mpParent->ReadControl<false>(ml, addr);
		};

		handlers.mpWriteHandler = [](void *thisptr, uint32 addr, uint8 value) -> bool {
			MemoryLayer& ml = *(MemoryLayer *)thisptr;

			return ml.mpParent->WriteControl(ml, addr, value);
		};

		ml.mpPhysLayer = mpMemMan->CreateLayer(kATMemoryPri_Cartridge1, handlers, address >> 8, size >> 8);

		mpMemMan->SetLayerModes(ml.mpPhysLayer, (ATMemoryAccessMode)memoryLayerMode);
	} else
		throw ATInvalidCustomDeviceDescException(objectNode, "Memory layer type not specified.");

	if (const VDJSONValueRef& cartModeRef = mlMembers.Optional("cart_mode"); cartModeRef.IsValid()) {
		const VDStringSpanW cartMode(ParseRequiredString(cartModeRef));

		if (cartMode == L"left") {
			ml.mbAutoRD5 = true;
			ml.mbRD5Active = true;
		} else if (cartMode == L"right")
			ml.mbAutoRD4 = true;
		else if (cartMode == L"cctl")
			ml.mbAutoCCTL = true;
		else if (cartMode == L"auto") {
			if (address >= 0xA000 && address < 0xC000 && 0xC000 - address <= size) {
				ml.mbAutoRD5 = true;
				ml.mbRD5Active = true;
			} else if (address >= 0x8000 && address < 0xA000 && 0xA000 - address <= size) {
				ml.mbAutoRD4 = true;
			} else if (address >= 0xD500 && address < 0xD600 && 0xD600 - address <= size) {
				ml.mbAutoCCTL = true;
			} else
				throw ATInvalidCustomDeviceDescException(cartModeRef, "Cannot use 'auto' mode as memory layer address range does not map to a cartridge region.");
		} else
			throw ATInvalidCustomDeviceDescException(cartModeRef, "Invalid cartridge mode.");
	}

	VDStringW name(mlMembers.RequiredString("name"));

	// Currently the layer name is used by the debugger, which does not expect localized
	// names -- so filter out loc chars for now. We also need to copy the name as the memory
	// manager expects a static name.
	size_t nameLen = name.size();
	char *name8 = (char *)mConfigAllocator.Allocate(nameLen + 1);

	for(size_t i = 0; i < nameLen; ++i) {
		wchar_t c = name[i];

		if (c < 0x20 || c >= 0x7F)
			c = L'_';

		name8[i] = (char)c;
	}

	name8[nameLen] = 0;

	if (ml.mpPhysLayer)
		mpMemMan->SetLayerName(ml.mpPhysLayer, name8);

	mlMembers.AssertNoUnused();
}

void ATDeviceCustom::ProcessSIODevices(VDJSONValueRef nodeRef) {
	if (!nodeRef.IsArray())
		throw ATInvalidCustomDeviceDescException(nodeRef, "SIO entry must be an array.");

	mpSIODeviceTable = mConfigAllocator.Allocate<SIODeviceTable>();

	for(VDJSONValueRef deviceNode : nodeRef.AsArray()) {
		ProcessSIODevice(deviceNode);
	}
}

void ATDeviceCustom::ProcessSIODevice(VDJSONValueRef nodeRef) {
	if (!nodeRef.IsObject())
		throw ATInvalidCustomDeviceDescException(nodeRef, "SIO device entry must be an object.");

	MemberParser devMembers(nodeRef);

	uint32 deviceId = devMembers.RequiredUint32("device_id");
	if (deviceId >= 0x100)
		throw ATInvalidCustomDeviceDescException(nodeRef, "Invalid SIO device ID.");

	const auto& deviceCountNode = devMembers.Optional("device_count");
	uint32 deviceIdCount = deviceCountNode.IsValid() ? ParseRequiredUint32(deviceCountNode) : 1;

	if (0x100 - deviceId < deviceIdCount)
		throw ATInvalidCustomDeviceDescException(nodeRef, "Invalid SIO device ID range.");

	SIOCommandTable *cmdTable = mConfigAllocator.Allocate<SIOCommandTable>();

	for(uint32 i=0; i<deviceIdCount; ++i) {
		if (mpSIODeviceTable->mpDevices[deviceId + i])
			throw ATInvalidCustomDeviceDescException(nodeRef, "SIO device already defined.");

		mpSIODeviceTable->mpDevices[deviceId + i] = cmdTable;
	}

	const auto& commandsNode = devMembers.Required("commands");
	if (!commandsNode.IsArray())
		throw ATInvalidCustomDeviceDescException(nodeRef, "SIO device requires commands array.");

	for(const VDJSONValueRef& commandNode : commandsNode.AsArray()) {
		MemberParser commandMembers(commandNode);
		const auto& commandIdNode = commandMembers.Required("id");
		bool isCopy = false;
		uint32 commandId = 0;

		SIOCommand *cmd;
		if (const VDJSONValueRef& copyFromNode = commandMembers.Optional("copy_from"); copyFromNode.IsValid()) {
			uint32 srcId = ParseRequiredUint32(copyFromNode);

			if (srcId >= 0x100)
				throw ATInvalidCustomDeviceDescException(copyFromNode, "Invalid source command ID.");

			cmd = cmdTable->mpCommands[srcId];
			if (!cmd)
				throw ATInvalidCustomDeviceDescException(copyFromNode, "Source command is not defined.");

			isCopy = true;
		} else {
			cmd = mConfigAllocator.Allocate<SIOCommand>();
		}

		if (commandIdNode.IsString() && !wcscmp(commandIdNode.AsString(), L"default")) {
			for(SIOCommand*& cmdEntry : cmdTable->mpCommands) {
				if (!cmdEntry)
					cmdEntry = cmd;
			}
		} else {
			commandId = ParseRequiredUint32(commandIdNode);

			if (commandId >= 0x100)
				throw ATInvalidCustomDeviceDescException(commandIdNode, "Invalid SIO command ID.");

			if (cmdTable->mpCommands[commandId])
				throw ATInvalidCustomDeviceDescException(commandIdNode, "Conflicting command ID in SIO device command list.");

			cmdTable->mpCommands[commandId] = cmd;
		}

		if (isCopy)
			continue;

		if (const VDJSONValueRef& autoTransferNode = commandMembers.Optional("auto_transfer"); autoTransferNode.IsValid()) {
			MemberParser autoTransferMembers(autoTransferNode);

			uint32 id = autoTransferNode.GetDocument()->mNameTable.GetToken(autoTransferMembers.Required("segment").AsString()).mToken;
			Segment *segment = nullptr;

			for(Segment& seg : mSegments) {
				if (seg.mId == id) {
					segment = &seg;
					break;
				}
			}

			if (!segment)
				throw ATInvalidCustomDeviceDescException(autoTransferNode, "Unknown segment name.");

			const VDStringSpanW mode(ParseRequiredString(autoTransferMembers.Required("mode")));
			if (mode == L"read") {
				cmd->mbAutoTransferWrite = false;
			} else if (mode == L"write") {
				cmd->mbAutoTransferWrite = true;
			} else
				throw ATInvalidCustomDeviceDescException(autoTransferNode, "Unknown auto-transfer mode.");

			const auto& offsetNode = autoTransferMembers.Optional("offset");
			const uint32 offset = offsetNode.IsValid() ? ParseRequiredUint32(offsetNode) : 0;
			const uint32 length = autoTransferMembers.RequiredUint32("length");

			if (length < 1 || length > 8192)
				throw ATInvalidCustomDeviceDescException(autoTransferNode, "Invalid transfer length (must be 1-8192 bytes).");

			if (offset >= segment->mSize || segment->mSize - offset < length)
				throw ATInvalidCustomDeviceDescException(autoTransferNode, "Offset/length specifies range outside of segment.");

			cmd->mpAutoTransferSegment = segment;
			cmd->mAutoTransferOffset = offset;
			cmd->mAutoTransferLength = length;
		} else {
			cmd->mpScript = ParseScriptOpt(kATCDVMTypeVoid, commandMembers.Optional("script"));
		}

		commandMembers.AssertNoUnused();
	}
}

void ATDeviceCustom::ProcessControllerPorts(VDJSONValueRef nodeRef) {
	if (!nodeRef.IsObject())
		throw ATInvalidCustomDeviceDescException(nodeRef, "controller_ports member must be an object.");

	uint32 boundControllers = 0;

	for(const auto& controllerEntry : nodeRef.AsObject()) {
		uint32 portIndex = ParseRequiredUint32(controllerEntry.GetValue());

		if (portIndex >= 4)
			throw ATInvalidCustomDeviceDescException(controllerEntry.GetValue(), "Invalid controller port index.");

		const uint32 bit = 1 << portIndex;

		if (boundControllers & bit)
			throw ATInvalidCustomDeviceDescException(controllerEntry.GetValue(), "Controller port is already bound.");

		boundControllers |= bit;

		if (!mpCompiler->DefineObjectVariable(VDTextWToU8(VDStringSpanW(controllerEntry.GetName())).c_str(), &mControllerPorts[portIndex], ControllerPort::kVMObjectClass))
			throw ATInvalidCustomDeviceDescException(controllerEntry.GetValue(), mpCompiler->GetError());

		mControllerPorts[portIndex].Init();
	}
}

const ATCDVMFunction *ATDeviceCustom::ParseScriptOpt(const ATCDVMTypeInfo& returnType, VDJSONValueRef valueRef) {
	if (!valueRef.IsValid())
		return nullptr;

	return ParseScript(returnType, valueRef);
}

const ATCDVMFunction *ATDeviceCustom::ParseScript(const ATCDVMTypeInfo& returnType, VDJSONValueRef valueRef) {
	VDStringA buf;
	vdfastvector<std::pair<size_t, VDJSONValueRef>> lineLookup;

	if (valueRef.IsString()) {
		lineLookup.push_back({0, valueRef});
		buf = VDTextWToU8(VDStringSpanW(valueRef.AsString()));
		buf += '\n';
	} else if (valueRef.IsArray()) {
		for(VDJSONValueRef arrayEl : valueRef.AsArray()) {
			if (!arrayEl.IsString())
				throw ATInvalidCustomDeviceDescException(valueRef, "Script must be a string or array of strings.");
		
			lineLookup.push_back({buf.size(), arrayEl});

			buf += VDTextWToU8(VDStringSpanW(arrayEl.AsString()));
			buf += '\n';
		}
	} else
		throw ATInvalidCustomDeviceDescException(valueRef, "Script must be a string or array of strings.");

	const ATCDVMFunction *func = mpCompiler->CompileProcedure(returnType, buf.data(), buf.size());

	if (!func) {
		const auto& failNode = std::upper_bound(lineLookup.begin(), lineLookup.end(), mpCompiler->GetErrorOffset(),
			[](uint32 x, const auto& y) { return x < y.first; })[-1].second;

		throw ATInvalidCustomDeviceDescException(failNode, mpCompiler->GetError());
	}

	return func;
}

vdvector_view<uint8> ATDeviceCustom::ParseBlob(VDJSONValueRef valueRef) {
	uint32 size = 1;
	if (valueRef.IsArray()) {
		size = valueRef.GetArrayLength();

		if (!size)
			throw ATInvalidCustomDeviceDescException(valueRef, "Array cannot be empty.");
	} else if (valueRef.IsString()) {
		size = wcslen(valueRef.AsString());
		if (!size)
			throw ATInvalidCustomDeviceDescException(valueRef, "String cannot be empty.");
	}

	uint8 *data = (uint8 *)mConfigAllocator.Allocate(size, 4);

	if (valueRef.IsArray()) {
		for(uint32 i = 0; i < size; ++i)
			data[i] = ParseRequiredUint8(valueRef[i]);
	} else if (valueRef.IsString()) {
		const wchar_t *s = valueRef.AsString();

		for(uint32 i = 0; i < size; ++i) {
			wchar_t c = s[i];

			if (c < 0x20 || c >= 0x7F)
				throw ATInvalidCustomDeviceDescException(valueRef, "String data must be ASCII.");

			data[i] = (uint8)c;
		}
	} else
		data[0] = ParseRequiredUint8(valueRef);

	return vdvector_view(data, size);
}

uint8 ATDeviceCustom::ParseRequiredUint8(VDJSONValueRef valueRef) {
	uint32 v = ParseRequiredUint32(valueRef);

	if (v != (uint8)v)
		throw ATInvalidCustomDeviceDescException(valueRef, "Value out of range.");

	return (uint8)v;
}

uint32 ATDeviceCustom::ParseRequiredUint32(VDJSONValueRef valueRef) {
	if (valueRef.IsInt()) {
		sint64 v = valueRef.AsInt64();

		if (v != (uint32)v)
			throw ATInvalidCustomDeviceDescException(valueRef, "Value out of range.");

		return (uint32)v;
	} else if (valueRef.IsReal()) {
		double v = valueRef.AsDouble();

		if (v < 0 || v >= 4294967296.0 || v != (uint32)v)
			throw ATInvalidCustomDeviceDescException(valueRef, "Value out of range.");

		return (uint32)v;
	} else if (valueRef.IsString()) {
		return ParseStringAsUint32(valueRef.AsString(), valueRef);
	} else
		throw ATInvalidCustomDeviceDescException(valueRef, "Value out of range.");
}

uint32 ATDeviceCustom::ParseStringAsUint32(const wchar_t *s, VDJSONValueRef context) {
	wchar_t *t = nullptr;
	unsigned long v;

	if (s[0] == '$')
		v = wcstoul(s + 1, &t, 16);
	else if (s[0] == '0' && (s[1] == 'x' || s[1] == 'X'))
		v = wcstoul(s + 2, &t, 16);
	else
		v = wcstoul(s, &t, 10);

	if (*t || v != (uint32)v)
		throw ATInvalidCustomDeviceDescException(context, "Value out of range.");

	return (uint32)v;
}

const wchar_t *ATDeviceCustom::ParseRequiredString(VDJSONValueRef valueRef) {
	if (!valueRef.IsString())
		throw ATInvalidCustomDeviceDescException(valueRef, "Value out of range.");

	return valueRef.AsString();
}

std::optional<bool> ATDeviceCustom::ParseBool(VDJSONValueRef valueRef) {
	if (!valueRef.IsValid())
		return {};
	else if (valueRef.IsBool())
		return valueRef.AsBool();
	else
		throw ATInvalidCustomDeviceDescException(valueRef, "Value out of range.");
}

void ATDeviceCustom::LoadDependency(VDJSONValueRef valueRef, ATVFSFileView **view) {
	const wchar_t *name = ParseRequiredString(valueRef);

	VDStringW path(name);
	VDStringW basePath;
	VDStringW subPath;
	for(;;) {
		auto pathType = ATParseVFSPath(path.c_str(), basePath, subPath);
		if (pathType == kATVFSProtocol_None)
			break;

		path = basePath;

		if (pathType == kATVFSProtocol_File)
			break;
	}

	if (wcschr(path.c_str(), '/') || wcschr(path.c_str(), '\\') || wcschr(path.c_str(), ':') || wcschr(path.c_str(), '%'))
		throw ATInvalidCustomDeviceDescException(valueRef, "Source file must be a local filename with no directory component.");

	OpenViewWithTracking(ATMakeVFSPath(mResourceBasePath.c_str(), name).c_str(), view);
}

void ATDeviceCustom::ClearFileTracking() {
	mTrackedFiles.clear();
}

void ATDeviceCustom::OpenViewWithTracking(const wchar_t *path, ATVFSFileView **view) {
	if (mbHotReload) {
		const wchar_t *path2 = path;

		VDStringW basePath, subPath, prevPath;
		VDFile f;

		for(;;) {
			auto protocol = ATParseVFSPath(path2, basePath, subPath);

			if (protocol == kATVFSProtocol_None)
				break;

			if (protocol == kATVFSProtocol_File) {
				FileTrackingInfo fti {};

				if (f.openNT(path2)) {
					fti.mSize = (uint64)f.size();
					fti.mTimestamp = f.getLastWriteTime().mTicks;
				}

				// store an entry even if the file open fails, as we still want to retry if the file
				// becomes readable
				mTrackedFiles.insert_as(path).first->second = fti;

				// leave the file open so the read lock persists, and maybe we can save some file open overhead
				break;
			}

			prevPath.swap(basePath);
			path2 = prevPath.c_str();
		}
	}

	ATVFSOpenFileView(path, false, view);
}

bool ATDeviceCustom::CheckForTrackedChanges() {
	if (!mbHotReload)
		return false;

	for(const auto& trackedFile : mTrackedFiles) {
		VDFile f;
		uint64 fsize = 0;
		uint64 timestamp = 0;

		if (f.openNT(trackedFile.first.c_str())) {
			fsize = (uint64)f.size();
			timestamp = f.getLastWriteTime().mTicks;
		}

		if (fsize != trackedFile.second.mSize || timestamp != trackedFile.second.mTimestamp)
			return true;
	}

	return false;
}

///////////////////////////////////////////////////////////////////////////

void ATDeviceCustom::Segment::VMCallClear(sint32 value) {
	if (mbReadOnly)
		return;

	memset(mpData, value, mSize);
}

void ATDeviceCustom::Segment::VMCallFill(sint32 offset, sint32 value, sint32 size) {
	if (offset < 0 || size <= 0 || mbReadOnly)
		return;

	if ((uint32)offset >= mSize || mSize - (uint32)offset < (uint32)size)
		return;

	memset(mpData + offset, value, size);
}

void ATDeviceCustom::Segment::VMCallCopy(sint32 destOffset, Segment *srcSegment, sint32 srcOffset, sint32 size) {
	if (destOffset < 0 || srcOffset < 0 || size <= 0 || mbReadOnly)
		return;

	uint32 udoffset = (uint32)destOffset;
	uint32 usoffset = (uint32)srcOffset;
	uint32 usize = (uint32)size;

	if (udoffset >= mSize || mSize - udoffset < usize)
		return;

	if (usoffset >= srcSegment->mSize || srcSegment->mSize - usoffset < usize)
		return;

	memmove(mpData + udoffset, srcSegment->mpData + usoffset, usize);
}

void ATDeviceCustom::Segment::VMCallWriteByte(sint32 offset, sint32 value) {
	if (offset >= 0 && (uint32)offset < mSize && !mbReadOnly)
		mpData[offset] = (uint8)value;
}

sint32 ATDeviceCustom::Segment::VMCallReadByte(sint32 offset) {
	if (offset >= 0 && (uint32)offset < mSize)
		return mpData[offset];
	else
		return 0;
}

///////////////////////////////////////////////////////////////////////////

void ATDeviceCustom::MemoryLayer::VMCallSetOffset(sint32 offset) {
	if (mpPhysLayer && mpSegment) {
		if (!(offset & 0xFF) && offset >= 0 && (uint32)offset <= mMaxOffset)
			mpParent->mpMemMan->SetLayerMemory(mpPhysLayer, mpSegment->mpData + offset);
	}
}

void ATDeviceCustom::MemoryLayer::VMCallSetSegmentAndOffset(Segment *seg, sint32 offset) {
	if (!seg)
		return;

	if (mpPhysLayer && !(offset & 0xFF) && offset >= 0 && (uint32)offset < seg->mSize && seg->mSize - (uint32)offset >= mSize) {
		mpSegment = seg;
		mMaxOffset = seg->mSize - mSize;

		mpParent->mpMemMan->SetLayerMemory(mpPhysLayer, seg->mpData + offset);
	}
}

void ATDeviceCustom::MemoryLayer::VMCallSetModes(sint32 read, sint32 write) {
	if (!mpPhysLayer)
		return;

	uint32 modes = 0;

	if (write)
		modes |= kATMemoryAccessMode_W;

	if (read)
		modes |= kATMemoryAccessMode_AR;

	mpParent->mpMemMan->SetLayerModes(mpPhysLayer, (ATMemoryAccessMode)modes);

	if (mbAutoRD5) {
		bool rd5Active = (modes != 0);

		if (mbRD5Active != rd5Active) {
			mbRD5Active = rd5Active;
			mpParent->mpCartPort->OnLeftWindowChanged(mpParent->mCartId, mpParent->IsLeftCartActive());
		}
	}
}

void ATDeviceCustom::MemoryLayer::VMCallSetReadOnly(sint32 ro) {
	if (mpPhysLayer)
		mpParent->mpMemMan->SetLayerReadOnly(mpPhysLayer, ro != 0);
}

///////////////////////////////////////////////////////////////////////////

sint32 ATDeviceCustom::Network::VMCallSendMessage(sint32 param1, sint32 param2) {
	return mpParent->SendNetCommand(param1, param2, NetworkCommand::ScriptEvent);
}

///////////////////////////////////////////////////////////////////////////

void ATDeviceCustom::SIO::VMCallAck() {
	if (!mbValid)
		return;

	if (mbDataFrameReceived) {
		mbDataFrameReceived = false;

		// enforce data-to-ACK delay (850us minimum)
		mpParent->mpSIOMgr->Delay(1530);
	}

	mpParent->mpSIOMgr->SendACK();
}

void ATDeviceCustom::SIO::VMCallNak() {
	if (!mbValid)
		return;

	if (mbDataFrameReceived) {
		mbDataFrameReceived = false;

		// enforce data-to-NAK delay (850us minimum)
		mpParent->mpSIOMgr->Delay(1530);
	}

	mpParent->mpSIOMgr->SendNAK();
}

void ATDeviceCustom::SIO::VMCallError() {
	if (!mbValid)
		return;

	mpParent->mpSIOMgr->SendError();
}

void ATDeviceCustom::SIO::VMCallComplete() {
	if (!mbValid)
		return;

	mpParent->mpSIOMgr->SendComplete();
}

void ATDeviceCustom::SIO::VMCallSendFrame(Segment *seg, sint32 offset, sint32 length) {
	if (!mbValid)
		return;

	if (!seg || offset < 0 || length <= 0 || length > 8192)
		return;

	uint32 uoffset = (uint32)offset;
	uint32 ulength = (uint32)length;
	if (uoffset > seg->mSize || seg->mSize - uoffset < ulength)
		return;

	mpParent->mpSIOMgr->SendData(seg->mpData + offset, ulength, true);
	mpParent->mpSIOMgr->InsertFence((uint32)SerialFenceId::ScriptDelay);
	mpParent->mVMThreadSIO.Suspend();
}

void ATDeviceCustom::SIO::VMCallRecvFrame(sint32 length) {
	if (!mbValid)
		return;

	if (length <= 0 || length > 8192)
		return;

	mbDataFrameReceived = true;
	mpParent->mpSIOMgr->ReceiveData((uint32)SerialFenceId::ScriptReceive, (uint32)length, true);
	mpParent->mVMThreadSIO.Suspend();
}

void ATDeviceCustom::SIO::VMCallDelay(sint32 cycles) {
	if (!mbValid)
		return;

	if (cycles <= 0 || cycles > 64*1024*1024)
		return;

	mpParent->mpSIOMgr->Delay((uint32)cycles);
	mpParent->mpSIOMgr->InsertFence((uint32)SerialFenceId::ScriptDelay);
	mpParent->mVMThreadSIO.Suspend();
}

///////////////////////////////////////////////////////////////////////////

void ATDeviceCustom::Clock::Reset() {
	mLocalTimeCaptureTimestamp = 0;
	mLocalTime = {};
}

void ATDeviceCustom::Clock::VMCallCaptureLocalTime() {
	uint64 t = mpParent->mpScheduler->GetTick64();

	if (mLocalTimeCaptureTimestamp != t) {
		mLocalTimeCaptureTimestamp = t;

		mLocalTime = VDGetLocalDate(VDGetCurrentDate());
	}
}

sint32 ATDeviceCustom::Clock::VMCallLocalYear() {
	return mLocalTime.mYear;
}

sint32 ATDeviceCustom::Clock::VMCallLocalMonth() {
	return mLocalTime.mMonth;
}

sint32 ATDeviceCustom::Clock::VMCallLocalDay() {
	return mLocalTime.mDay;
}

sint32 ATDeviceCustom::Clock::VMCallLocalDayOfWeek() {
	return mLocalTime.mDayOfWeek;
}

sint32 ATDeviceCustom::Clock::VMCallLocalHour() {
	return mLocalTime.mHour;
}

sint32 ATDeviceCustom::Clock::VMCallLocalMinute() {
	return mLocalTime.mMinute;
}

sint32 ATDeviceCustom::Clock::VMCallLocalSecond() {
	return mLocalTime.mSecond;
}

///////////////////////////////////////////////////////////////////////////

void ATDeviceCustom::ControllerPort::Init() {
	if (mPortInput < 0) {
		mPortInput = mpPortController->AllocatePortInput(mbPort2, -1);

		mInputMask = 0;

		mPotA = 229;
		mPotB = 229;
	}
}

void ATDeviceCustom::ControllerPort::Shutdown() {
	if (mPortInput >= 0) {
		ResetPotPositions();
		mbPaddleASet = false;
		mbPaddleBSet = false;

		mpPortController->FreePortInput(mPortInput);
		mPortInput = -1;
	}
}

void ATDeviceCustom::ControllerPort::Enable() {
	if (!mbEnabled) {
		mbEnabled = true;

		Reapply();
	}
}

void ATDeviceCustom::ControllerPort::Disable() {
	if (mbEnabled) {
		mbEnabled = false;

		Reapply();
	}
}

void ATDeviceCustom::ControllerPort::Reapply() {
	if (mPortInput < 0)
		return;

	if (mbEnabled) {
		mpPortController->SetPortInput(mPortInput, mInputMask);

		if (mbPaddleASet)
			mpPortController->SetPotPosition(mbPort2 ? 2 : 0, mPotA);

		if (mbPaddleBSet)
			mpPortController->SetPotPosition(mbPort2 ? 3 : 1, mPotB);
	} else {
		mpPortController->SetPortInput(mPortInput, 0);
		ResetPotPositions();
	}
}

void ATDeviceCustom::ControllerPort::ResetPotPositions() {
	if (mbPaddleASet)
		mpPortController->SetPotPosition(mbPort2 ? 2 : 0, 229);

	if (mbPaddleBSet)
		mpPortController->SetPotPosition(mbPort2 ? 3 : 1, 229);
}

void ATDeviceCustom::ControllerPort::VMCallSetPaddleA(sint32 pos) {
	mbPaddleASet = true;

	if (pos >= 228)
		pos = 228;
	else if (pos < 1)
		pos = 1;

	if (mPotA != pos) {
		mPotA = pos;

		mpPortController->SetPotPosition(mbPort2 ? 2 : 0, mPotA);
	}
}

void ATDeviceCustom::ControllerPort::VMCallSetPaddleB(sint32 pos) {
	mbPaddleBSet = true;

	if (pos >= 228)
		pos = 228;
	else if (pos < 1)
		pos = 1;

	if (mPotB != pos) {
		mPotB = pos;

		mpPortController->SetPotPosition(mbPort2 ? 3 : 1, mPotB);
	}
}

void ATDeviceCustom::ControllerPort::VMCallSetTrigger(sint32 asserted) {
	uint32 newMask = mInputMask;

	if (asserted)
		newMask |= 0x100;
	else
		newMask &= ~UINT32_C(0x100);

	if (mInputMask != newMask) {
		mInputMask = newMask;

		if (mbEnabled)
			mpPortController->SetPortInput(mPortInput, mInputMask);
	}
}

void ATDeviceCustom::ControllerPort::VMCallSetDirs(sint32 mask) {
	uint32 newMask = mInputMask ^ ((mask ^ mInputMask) & 15);

	if (mInputMask != newMask) {
		mInputMask = newMask;

		if (mbEnabled)
			mpPortController->SetPortInput(mPortInput, mInputMask);
	}
}
