//	Altirra - Atari 800/800XL/5200 emulator
//	Copyright (C) 2009-2016 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include <stdafx.h>
#include <vd2/system/hash.h>
#include <vd2/system/int128.h>
#include <vd2/system/math.h>
#include <at/atcore/audiosource.h>
#include <at/atcore/logging.h>
#include <at/atcore/propertyset.h>
#include <at/atcore/deviceserial.h>
#include "audiosampleplayer.h"
#include "diskdrivepercom.h"
#include "memorymanager.h"
#include "firmwaremanager.h"
#include "debuggerlog.h"

extern ATLogChannel g_ATLCDiskEmu;

void ATCreateDeviceDiskDrivePercom(const ATPropertySet& pset, IATDevice **dev) {
	vdrefptr<ATDeviceDiskDrivePercom> p(new ATDeviceDiskDrivePercom);
	p->SetSettings(pset);

	*dev = p.release();
}

extern const ATDeviceDefinition g_ATDeviceDefDiskDrivePercom = { "diskdrivepercom", "diskdrivepercom", L"Percom disk drive (full emulation)", ATCreateDeviceDiskDrivePercom };

///////////////////////////////////////////////////////////////////////////

void ATDeviceDiskDrivePercom::Drive::OnDiskChanged(bool mediaRemoved) {
	mpParent->OnDiskChanged(mIndex, mediaRemoved);
}

void ATDeviceDiskDrivePercom::Drive::OnWriteModeChanged() {
	mpParent->OnWriteModeChanged(mIndex);
}

void ATDeviceDiskDrivePercom::Drive::OnTimingModeChanged() {
	mpParent->OnTimingModeChanged(mIndex);
}

void ATDeviceDiskDrivePercom::Drive::OnAudioModeChanged() {
	mpParent->OnAudioModeChanged(mIndex);
}

bool ATDeviceDiskDrivePercom::Drive::IsImageSupported(const IATDiskImage& image) const {
	return true;
}

///////////////////////////////////////////////////////////////////////////

ATDeviceDiskDrivePercom::ATDeviceDiskDrivePercom() {
	mBreakpointsImpl.BindBPHandler(mCoProc);
	mBreakpointsImpl.SetStepHandler(this);

	mSerialCmdQueue.SetOnDriveCommandStateChanged(
		[this](bool asserted) {
			mACIA.SetCTS(asserted);
		}
	);

	mTargetHistoryProxy.Init(mCoProc);
	InitTargetControl(mTargetHistoryProxy, 1000000.0, kATDebugDisasmMode_6809, &mBreakpointsImpl);

	mFirmwareControl.Init(mROM, sizeof mROM, kATFirmwareType_Percom);
}

ATDeviceDiskDrivePercom::~ATDeviceDiskDrivePercom() {
}

void *ATDeviceDiskDrivePercom::AsInterface(uint32 iid) {
	switch(iid) {
		case IATDeviceScheduling::kTypeID: return static_cast<IATDeviceScheduling *>(this);
		case IATDeviceFirmware::kTypeID: return static_cast<IATDeviceFirmware *>(&mFirmwareControl);
		case IATDeviceDiskDrive::kTypeID: return static_cast<IATDeviceDiskDrive *>(this);
		case IATDeviceSIO::kTypeID: return static_cast<IATDeviceSIO *>(this);
		case IATDeviceAudioOutput::kTypeID: return static_cast<IATDeviceAudioOutput *>(&mAudioPlayer);
		case ATFDCEmulator::kTypeID: return &mFDC;
	}

	return ATDiskDriveDebugTargetControl::AsInterface(iid);
}

void ATDeviceDiskDrivePercom::GetDeviceInfo(ATDeviceInfo& info) {
	info.mpDef = &g_ATDeviceDefDiskDrivePercom;
}

void ATDeviceDiskDrivePercom::GetSettingsBlurb(VDStringW& buf) {
	bool first = true;

	for(uint32 i=0; i<kNumDrives; ++i) {
		if (mDrives[i].mType) {
			if (first)
				first = false;
			else
				buf += ',';

			buf.append_sprintf(L"D%u:", i + mDriveId + 1);
		}
	}
}

void ATDeviceDiskDrivePercom::GetSettings(ATPropertySet& settings) {
	settings.SetUint32("id", mDriveId);

	VDStringA s;
	for(uint32 i=0; i<kNumDrives; ++i) {
		s.sprintf("drivetype%u", i);
		settings.SetUint32(s.c_str(), (uint32)mDrives[i].mType);
	}
}

bool ATDeviceDiskDrivePercom::SetSettings(const ATPropertySet& settings) {
	VDStringA s;
	bool change = false;

	for(uint32 i=0; i<kNumDrives; ++i) {
		s.sprintf("drivetype%u", i);
		const uint32 driveTypeCode = settings.GetUint32(s.c_str(), i ? kDriveType_None : kDriveType_5_25_40Track);

		if (driveTypeCode <= kDriveType_5_25_80Track) {
			Drive& drive = mDrives[i];

			if (drive.mType != driveTypeCode) {
				drive.mType = (DriveType)driveTypeCode;

				if (drive.mType == kDriveType_5_25_80Track)
					drive.mMaxTrack = 180;
				else
					drive.mMaxTrack = 90;

				change = true;
			}
		}
	}

	uint32 newDriveId = settings.GetUint32("id", mDriveId) & 7;

	if (mDriveId != newDriveId) {
		mDriveId = newDriveId;
		change = true;
	}

	return !change;
}

void ATDeviceDiskDrivePercom::Init() {
	mSerialCmdQueue.Init(&mDriveScheduler, mpSIOMgr);
	mSerialXmitQueue.Init(mpScheduler, mpSIOMgr);

	uintptr *readmap = mCoProc.GetReadMap();
	uintptr *writemap = mCoProc.GetWriteMap();

	ATCoProcMemoryMapView mmap(readmap, writemap);

	// initialize memory map
	mmap.Clear(mDummyRead, mDummyWrite);

	// Map hardware registers to $D000-D3FF. The individual sections are selected
	// by A2-A5, so all page mappings are the same here.
	mReadNodeHardware.mpThis = this;
	mReadNodeHardware.mpRead = [](uint32 addr, void *thisPtr) { return ((ATDeviceDiskDrivePercom *)thisPtr)->OnHardwareRead(addr); };
	mReadNodeHardware.mpDebugRead = [](uint32 addr, void *thisPtr) { return ((ATDeviceDiskDrivePercom *)thisPtr)->OnHardwareDebugRead(addr); };

	mWriteNodeHardware.mpThis = this;
	mWriteNodeHardware.mpWrite = [](uint32 addr, uint8 value, void *thisPtr) { ((ATDeviceDiskDrivePercom *)thisPtr)->OnHardwareWrite(addr, value); };

	mmap.SetReadHandler(0xD0, 0x04, mReadNodeHardware);
	mmap.SetWriteHandler(0xD0, 0x04, mWriteNodeHardware);

	// map RAM to $DC00-DFFF
	mmap.SetMemory(0xDC, 0x04, mRAM);

	// map ROM to $F000-FFFF (mirrored)
	mmap.SetReadMem(0xF0, 0x08, mROM);
	mmap.SetReadMem(0xF8, 0x08, mROM);

	mDriveScheduler.SetRate(VDFraction(1000000, 1));

	// Base clock to the ACIA is 4MHz / 13.
	mACIA.Init(&mDriveScheduler);
	mACIA.SetMasterClockPeriod(13 * 16);
	mACIA.SetTransmitFn([this](uint8 v, uint32 cyclesPerBit) { OnACIATransmit(v, cyclesPerBit); });

	mFDC.Init(&mDriveScheduler, 288.0f, 2.0f, ATFDCEmulator::kType_2793);
	mFDC.SetAutoIndexPulse(true);
	mFDC.SetOnDrqChange([this](bool drq) { OnFDCDataRequest(drq); });
	mFDC.SetOnIrqChange([this](bool irq) { OnFDCInterruptRequest(irq); });
	mFDC.SetOnStep([this](bool inward) { OnFDCStep(inward); });

	int driveIndex = 0;
	for(auto& drive : mDrives) {
		drive.mDiskChangeHandler.Init(mpScheduler);
		drive.mDiskChangeHandler.SetOutputStateFns(
			[&drive, driveIndex, this](std::optional<bool> wpState) {
				if (mSelectedDrive == driveIndex)
					mFDC.SetWriteProtectOverride(wpState);
			},
			[&drive, driveIndex, this](std::optional<bool> readyState) {
				if (mSelectedDrive == driveIndex)
					mFDC.SetDiskImageReady(readyState);
			}
		);

		drive.OnDiskChanged(false);

		drive.OnWriteModeChanged();
		drive.OnTimingModeChanged();
		drive.OnAudioModeChanged();

		++driveIndex;
	}

	UpdateRotationStatus();
}

void ATDeviceDiskDrivePercom::Shutdown() {
	mAudioPlayer.Shutdown();
	mSerialCmdQueue.Shutdown();
	mSerialXmitQueue.Shutdown();

	ShutdownTargetControl();
	mFirmwareControl.Shutdown();

	if (mpSIOMgr) {
		mpSIOMgr->RemoveRawDevice(this);
		mpSIOMgr = nullptr;
	}

	for(auto& drive : mDrives) {
		drive.mDiskChangeHandler.Shutdown();

		if (drive.mpDiskInterface) {
			drive.mpDiskInterface->RemoveClient(&drive);
			drive.mpDiskInterface = nullptr;
		}
	}

	mpDiskDriveManager = nullptr;
}

uint32 ATDeviceDiskDrivePercom::GetComputerPowerOnDelay() const {
	return 20;
}

void ATDeviceDiskDrivePercom::WarmReset() {
}

void ATDeviceDiskDrivePercom::ComputerColdReset() {
	WarmReset();
}

void ATDeviceDiskDrivePercom::PeripheralColdReset() {
	memset(mRAM, 0xA5, sizeof mRAM);

	mACIA.Reset();
	mFDC.Reset();

	mSerialCmdQueue.Reset();
	mSerialXmitQueue.Reset();

	mbNmiState = false;
	mbNmiTimeout = false;
	mbNmiTimeoutEnabled = false;
	mDriveScheduler.UnsetEvent(mpEventDriveTimeout);

	ResetTargetControl();

	SelectDrive(-1);
	
	mbForcedIndexPulse = false;

	// start the disk drive on a track other than 0/20/39, just to make things interesting
	for(Drive& drive : mDrives) {
		drive.mCurrentTrack = 20;
		drive.mDiskChangeHandler.Reset();
	}

	mFDC.SetCurrentTrack(20, false);

	mbMotorRunning = false;
	mFDC.SetMotorRunning(false);
	mFDC.SetDensity(false);
	mFDC.SetWriteProtectOverride(false);
	mFDC.SetAutoIndexPulse(true);
	mFDC.SetSide(false);

	mCoProc.ColdReset();

	WarmReset();
}

void ATDeviceDiskDrivePercom::InitDiskDrive(IATDiskDriveManager *ddm) {
	mpDiskDriveManager = ddm;
	mAvailableDrives = 0;

	for(uint32 i=0; i<kNumDrives; ++i) {
		Drive& drive = mDrives[i];

		drive.mIndex = i;
		drive.mpParent = this;

		if (drive.mType) {
			drive.mpDiskInterface = ddm->GetDiskInterface(i + mDriveId);
			drive.mpDiskInterface->AddClient(&drive);

			mAvailableDrives |= (1 << i);
		}
	}
}

ATDeviceDiskDriveInterfaceClient ATDeviceDiskDrivePercom::GetDiskInterfaceClient(uint32 index) {
	for(uint32 i=0; i<kNumDrives; ++i) {
		if (mDrives[i].mType && !index--)
			return { &mDrives[i], mDriveId + i };
	}

	return {};
}

void ATDeviceDiskDrivePercom::InitSIO(IATDeviceSIOManager *mgr) {
	mpSIOMgr = mgr;
	mpSIOMgr->AddRawDevice(this);
}

void ATDeviceDiskDrivePercom::OnScheduledEvent(uint32 id) {
	if (id == kEventId_DriveTimeout) {
		mpEventDriveTimeout = nullptr;

		if (!mbNmiTimeout) {
			mbNmiTimeout = true;

			UpdateNmi();
		}
	} else
		return ATDiskDriveDebugTargetControl::OnScheduledEvent(id);
}

void ATDeviceDiskDrivePercom::OnCommandStateChanged(bool asserted) {
	if (mbCommandState != asserted) {
		mbCommandState = asserted;

		// Convert computer time to device time.
		//
		// We have a problem here because transmission is delayed by a byte time but we don't
		// necessarily know that delay when the command line is dropped. The XF551 has strict
		// requirements for the command line pulse because it needs about 77 machine cycles
		// from command asserted to start bit, but more importantly requires it to still be
		// asserted after the end of the last byte. To solve this, we assert /COMMAND
		// immediately but stretch the deassert a bit.

		const uint32 commandLatency = asserted ? 0 : 400;

		mSerialCmdQueue.AddCommandEdge(MasterTimeToDriveTime() + commandLatency, asserted);
	}
}

void ATDeviceDiskDrivePercom::OnMotorStateChanged(bool asserted) {
}

void ATDeviceDiskDrivePercom::OnReceiveByte(uint8 c, bool command, uint32 cyclesPerBit) {
	Sync();

	mACIA.ReceiveByte(c, (cyclesPerBit * 100 + (179/2)) / 179);
}

void ATDeviceDiskDrivePercom::OnSendReady() {
}

void ATDeviceDiskDrivePercom::OnDiskChanged(uint32 index, bool mediaRemoved) {
	Drive& drive = mDrives[index];

	if (mediaRemoved)
		drive.mDiskChangeHandler.ChangeDisk();

	UpdateDiskStatus();
}

void ATDeviceDiskDrivePercom::OnWriteModeChanged(uint32 index) {
	// FDC polls write protect state directly, so no action needed
}

void ATDeviceDiskDrivePercom::OnTimingModeChanged(uint32 index) {
	if (mSelectedDrive == (int)index) {
		const bool accurateTiming = mDrives[index].mpDiskInterface->IsAccurateSectorTimingEnabled();

		mFDC.SetAccurateTimingEnabled(accurateTiming);
	}
}

void ATDeviceDiskDrivePercom::OnAudioModeChanged(uint32 index) {
	if (mSelectedDrive == (int)index) {
		bool driveSounds = mDrives[index].mpDiskInterface->AreDriveSoundsEnabled();

		mbSoundsEnabled = driveSounds;

		UpdateRotationStatus();
	}
}

void ATDeviceDiskDrivePercom::Sync() {
	const uint32 limit = AccumSubCycles();

	for(;;) {
		if (!mCoProc.GetCyclesLeft()) {
			if (ATSCHEDULER_GETTIME(&mDriveScheduler) - limit < UINT32_C(0x80000000))
				break;

			ATSCHEDULER_ADVANCE(&mDriveScheduler);

			mCoProc.AddCycles(1);
		}

		mCoProc.Run();

		if (mCoProc.GetCyclesLeft()) {
			ScheduleImmediateResume();
			break;
		}
	}

	FlushStepNotifications();
}

uint8 ATDeviceDiskDrivePercom::OnHardwareDebugRead(uint32 addr) {
	// To access the hardware registers, A4 must be high and A5 must be low.
	// A2-A3 select the unit and A0-A1 the subunit (ACIA/FDC). A6-A9 don't
	// matter.
	switch(addr & 0x3C) {
		case 0x10:
			return mFDC.DebugReadByte(addr);

		case 0x14:
			return (mDriveId & 3) + (mDriveId & 4 ? 0x10 : 0x00) + 0xE0;

		case 0x30:
			return mACIA.DebugReadByte(addr);
	}

	return 0xFF;
}

uint8 ATDeviceDiskDrivePercom::OnHardwareRead(uint32 addr) {
	// To access the hardware registers, A4 must be high and A5 must be low.
	// A2-A3 select the unit and A0-A1 the subunit (ACIA/FDC). A6-A9 don't
	// matter.
	switch(addr & 0x3C) {
		case 0x10:
			if (mbNmiTimeoutEnabled)
				SetNmiTimeout();
			return ~mFDC.ReadByte(addr);

		case 0x14:
			return OnHardwareDebugRead(addr);

		case 0x30:
			return mACIA.ReadByte(addr);
	}

	return 0xFF;
}

void ATDeviceDiskDrivePercom::OnHardwareWrite(uint32 addr, uint8 value) {
	// To access the hardware registers, A4 must be high and A5 must be low.
	// A2-A3 select the unit and A0-A1 the subunit (ACIA/FDC). A6-A9 don't
	// matter.
	switch(addr & 0x3C) {
		case 0x10:
			mFDC.WriteByte(addr, ~value);
			if (mbNmiTimeoutEnabled)
				SetNmiTimeout();
			break;

		case 0x14:
			if (value & 8) {
				SetMotorEnabled(true);
				SelectDrive((value >> 1) & 3);
			} else {
				SetMotorEnabled(false);
				SelectDrive(-1);
			}

			mFDC.SetSide((value & 1) != 0);
			break;

		case 0x18: {
			const bool nmiTimeoutEnabled = (value & 1) != 0;

			if (mbNmiTimeoutEnabled != nmiTimeoutEnabled) {
				mbNmiTimeoutEnabled = nmiTimeoutEnabled;

				if (!mbNmiTimeoutEnabled)
					mDriveScheduler.UnsetEvent(mpEventDriveTimeout);

				UpdateNmi();
			}

			mFDC.SetDensity((value & 4) != 0);
			mFDC.SetAutoIndexPulse((value & 8) == 0);
			break;
		}

		case 0x30:
			mACIA.WriteByte(addr, value);
			break;
	}
}

void ATDeviceDiskDrivePercom::OnFDCDataRequest(bool asserted) {
	if (asserted)
		mCoProc.AssertFirq();
	else
		mCoProc.NegateFirq();
}

void ATDeviceDiskDrivePercom::OnFDCInterruptRequest(bool asserted) {
	UpdateNmi();
}

void ATDeviceDiskDrivePercom::OnFDCStep(bool inward) {
	if (mSelectedDrive < 0)
		return;

	Drive& drive = mDrives[mSelectedDrive];

	if (inward) {
		// step in (increasing track number)
		if (drive.mCurrentTrack < drive.mMaxTrack) {
			drive.mCurrentTrack += 2;

			mFDC.SetCurrentTrack(drive.mCurrentTrack, drive.mCurrentTrack == 0);
		}

		PlayStepSound();
	} else {
		// step out (decreasing track number)
		if (drive.mCurrentTrack > 0) {
			drive.mCurrentTrack -= 2;

			mFDC.SetCurrentTrack(drive.mCurrentTrack, drive.mCurrentTrack == 0);

			PlayStepSound();
		}
	}
}

void ATDeviceDiskDrivePercom::OnACIATransmit(uint8 v, uint32 cyclesPerBit) {
	mSerialXmitQueue.AddTransmitByte(DriveTimeToMasterTime() + mSerialXmitQueue.kTransmitLatency, v, (cyclesPerBit * 179 + 50) / 100);
}

void ATDeviceDiskDrivePercom::SetMotorEnabled(bool enabled) {
	if (mbMotorRunning != enabled) {
		mbMotorRunning = enabled;

		mFDC.SetMotorRunning(enabled);
		UpdateRotationStatus();
	}
}

void ATDeviceDiskDrivePercom::PlayStepSound() {
	if (!mbSoundsEnabled)
		return;

	const uint32 t = ATSCHEDULER_GETTIME(&mDriveScheduler);
	
	if (t - mLastStepSoundTime > 50000)
		mLastStepPhase = 0;

	mAudioPlayer.PlayStepSound(kATAudioSampleId_DiskStep2, 0.3f + 0.7f * cosf((float)mLastStepPhase++ * nsVDMath::kfPi));

	mLastStepSoundTime = t;
}

void ATDeviceDiskDrivePercom::UpdateRotationStatus() {
	if (mSelectedDrive >= 0) {
		const Drive& drive = mDrives[mSelectedDrive];

		drive.mpDiskInterface->SetShowMotorActive(mbMotorRunning);

		if (mbMotorRunning && mbSoundsEnabled) {
			mAudioPlayer.SetRotationSoundEnabled(true);
			return;
		}
	}

	mAudioPlayer.SetRotationSoundEnabled(false);
}

void ATDeviceDiskDrivePercom::UpdateDiskStatus() {
	if (mSelectedDrive >= 0) {
		const Drive& drive = mDrives[mSelectedDrive];
		IATDiskImage *image = drive.mpDiskInterface->GetDiskImage();

		mFDC.SetDiskImage(image);

		drive.mDiskChangeHandler.ForceOutputUpdate();
	} else
		mFDC.SetDiskImage(nullptr, false);
}

void ATDeviceDiskDrivePercom::SelectDrive(int index) {
	if (!(mAvailableDrives & (1 << index)))
		index = -1;

	if (mSelectedDrive == index)
		return;

	if (mSelectedDrive >= 0) {
		Drive& oldDrive = mDrives[mSelectedDrive];

		oldDrive.mpDiskInterface->SetShowMotorActive(false);
	}

	mSelectedDrive = index;

	if (index >= 0) {
		Drive& drive = mDrives[index];
		mFDC.SetDiskInterface(drive.mpDiskInterface);
		mFDC.SetCurrentTrack(drive.mCurrentTrack, drive.mCurrentTrack == 0);
		mFDC.SetSideMapping(ATFDCEmulator::SideMapping::Side2ReversedOffByOne, drive.mType == kDriveType_5_25_80Track ? 80 : 40);

		OnWriteModeChanged(index);
		OnTimingModeChanged(index);
		OnAudioModeChanged(index);
	} else {
		mFDC.SetDiskInterface(nullptr);
		mFDC.SetCurrentTrack(20, false);
	}

	UpdateDiskStatus();
	UpdateRotationStatus();
}

void ATDeviceDiskDrivePercom::UpdateNmi() {
	bool nmiState = true;

	// If the FDC is requesting an interrupt, the 74LS122 is kept in cleared
	// state pulling /NMI low. $DC14-17 bit 0 can also hold this low if set
	// to 0.
	if (!mFDC.GetIrqStatus() && mbNmiTimeoutEnabled) {
		// FDC is not requesting an interrupt and the NMI is enabled. See if
		// the timeout has expired.

		nmiState = mbNmiTimeout;
	}

	if (mbNmiState != nmiState) {
		mbNmiState = nmiState;

		if (nmiState)
			mCoProc.AssertNmi();
	}
}

void ATDeviceDiskDrivePercom::SetNmiTimeout() {
	// This timeout delay is determined by an R/C network connected to
	// the 74LS123. Not being able to interpret said circuit, use a timeout
	// that is enough for at least four disk rotations.
	const uint32 kTimeoutDelay = 1000000;

	mDriveScheduler.SetEvent(kTimeoutDelay, this, kEventId_DriveTimeout, mpEventDriveTimeout);

	// trigger pulse on /NMI
	mbNmiTimeout = false;
	UpdateNmi();
}
